"""
Base resources for SPICE package.
"""

from .rnn import BaseRNN
from .rnn_utils import DatasetRNN
from .rnn_training import fit_model
from .sindy_training import fit_spice
from .bandits import AgentNetwork, AgentSpice
from .sindy_utils import create_dataset, check_library_setup

__all__ = [
    'BaseRNN',
    'AgentNetwork',
    'AgentSpice',
    'create_dataset',
    'check_library_setup',
    'fit_model',
    'fit_spice',
    'DatasetRNN',
]


import optuna
import numpy as np

from .fit_sindy import fit_sindy_pipeline
from .bandits import AgentSpice, AgentNetwork, get_update_dynamics
from .sindy_utils import DatasetRNN
   

def optimize_for_participant(
    participant_id: int,
    agent_rnn: AgentNetwork,
    data: DatasetRNN,
    metric_rnn: float,
    rnn_modules: list,
    control_signals: list,
    library_setup: dict,
    filter_setup: dict,
    polynomial_degree: int,
    optimizer_type: str,
    n_sessions_off_policy: int,
    n_trials_off_policy: int,
    n_trials_same_action_off_policy: int,
    n_trials_optuna: int = 50,
    timeout: int = 600,  # ? 10 minutes timeout
    threshold: float = 0.01,
    verbose: bool = False,
):
    """
    Use Optuna to find the best optimizer type and hyperparameters for a specific participant.
    
    Args:
        variables: RNN module variables for the participant
        control: Control signals for the participant
        rnn_modules: List of RNN module names
        control_signals: List of control signal names
        library_setup: Dictionary mapping features to library components
        filter_setup: Dictionary mapping features to filter conditions
        polynomial_degree: Polynomial degree for the feature library
        n_trials: Number of Optuna trials to run
        timeout: Maximum time in seconds to run optimization
        verbose: Whether to print verbose output
        
    Returns:
        dict: Best optimizer configuration with type and parameters
    """
    
    threshold = 0.01
    
    def objective(trial):
        
        # Sample optimizer type
        # optimizer_type = trial.suggest_categorical("optimizer_type", ["STLSQ", "SR3_L1"])#, "SR3_weighted_l1"])
        
        # Sample optimizer hyperparameters
        optimizer_alpha = trial.suggest_float("optimizer_alpha", 0.01, 1.0, log=True)
        optimizer_threshold = trial.suggest_float("optimizer_threshold", 0.01, 0.2, log=True)
        
        # Sample off-policy parameters
        # n_sessions_off_policy = trial.suggest_categorical("n_sessions_off_policy", [0, 1])
        # n_trials_off_policy = trial.suggest_categorical("n_trials_off_policy", [1000, 2000])
        # n_trials_same_action_off_policy = trial.suggest_categorical("n_trials_same_action_off_policy", [5, 10, 20])
        
        # just fit the SINDy modules with the given parameters 
        sindy_modules = fit_sindy_pipeline(
            participant_id=participant_id,
            agent=agent_rnn,
            data=data,
            rnn_modules=rnn_modules,
            control_signals=control_signals,
            sindy_library_setup=library_setup,
            sindy_filter_setup=filter_setup,
            sindy_dataprocessing=None,
            optimizer_type=optimizer_type,
            optimizer_alpha=optimizer_alpha,
            optimizer_threshold=optimizer_threshold,
            polynomial_degree=polynomial_degree,
            shuffle=True,
            n_sessions_off_policy=n_sessions_off_policy,
            n_trials_off_policy=n_trials_off_policy,
            n_trials_same_action_off_policy=n_trials_same_action_off_policy,
            catch_convergence_warning=False,
            verbose=verbose,
        )
        
        spice_modules = {module: {} for module in sindy_modules}        
        for rnn_module in rnn_modules:
            spice_modules[rnn_module][participant_id] = sindy_modules[rnn_module]
        
        agent_spice = AgentSpice(model_rnn=agent_rnn._model, sindy_modules=spice_modules, n_actions=agent_rnn._n_actions)
        
        # compute loss
        probs_spice = get_update_dynamics(experiment=data.xs[0], agent=agent_spice)[1]
        
        # loss: Difference between average trial likelihoods of RNN and SPICE -> SPICE can become even better than RNN; But that does not make sense on off-policy data
        # lik_spice = np.exp(log_likelihood(data.xs[0, :probs_spice.shape[0], :agent_rnn._n_actions].numpy(), probs=probs_spice) / probs_spice.size)
        # loss = metric_rnn - lik_spice
        
        # loss: MSE between predicted trial probabilities
        loss_reconstruction = np.power(metric_rnn - probs_spice, 2).mean()
        # loss_parameter = bayesian_information_criterion(data.xs[0, :len(probs_spice), :agent_spice._n_actions].numpy(), probs_spice, n_parameters=agent_spice.count_parameters()[agent_spice.get_participant_ids()[0]])/len(probs_spice)
        penalty_parameters = 0
        penalty_parameters += np.sum([np.sum(np.abs(sindy_modules[module].coefficients())) for module in sindy_modules])
        loss = loss_reconstruction + penalty_parameters * 1e-4
        
        if loss == np.nan:
            loss = 1e3
                    
        return loss
    
    threshold_no_improvement = 50
    
    def callback_no_improvement(study, trial):
        """
        Callback to stop the study if there is no improvement over the last `patience` trials.

        Args:
            study: The Optuna study object.
            trial: The Optuna trial object.
            patience (int): Number of trials to wait for an improvement.
        """
        # Store the state in the study's `user_attrs`
        if "last_best_value" not in study.user_attrs:
            study.set_user_attr("last_best_value", study.best_value)
            study.set_user_attr("trial_count", 0)
            return

        last_best_value = study.user_attrs["last_best_value"]
        trial_count = study.user_attrs["trial_count"]

        # Check for improvement
        if study.best_value < last_best_value:
            study.set_user_attr("last_best_value", study.best_value)
            study.set_user_attr("trial_count", 0)
        else:
            trial_count += 1
            study.set_user_attr("trial_count", trial_count)

            # Stop the study if patience is exceeded
            if trial_count >= threshold_no_improvement:
                study.stop()
        
    def callback_threshold_reached(study, trial):
        try:
            if study.best_value < threshold:  # Use the threshold argument
                study.stop()
        except ValueError as e:
            if "No trials are completed yet" in str(e):
                pass
        
    study = optuna.create_study(direction="minimize")
    study.optimize(
        objective, 
        n_trials=n_trials_optuna, 
        timeout=timeout, 
        show_progress_bar=True, 
        callbacks=[
            callback_threshold_reached, 
            # callback_no_improvement,
            ],
    )
    
    return {
        # "optimizer_type": study.best_params["optimizer_type"],
        "optimizer_alpha": study.best_params["optimizer_alpha"],
        "optimizer_threshold": study.best_params["optimizer_threshold"],
        # "n_sessions_off_policy": study.best_params["n_sessions_off_policy"],
        # "n_trials_off_policy": study.best_params["n_trials_off_policy"],
        # "n_trials_same_action_off_policy": study.best_params["n_trials_same_action_off_policy"],
    }


import sys, os
from typing import List, Union, Dict

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
import matplotlib.cm as cm
import torch
from typing import Tuple, Optional

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from resources.bandits import AgentQ, AgentNetwork, AgentSpice, BanditSession, get_update_dynamics, plot_session as plt_session
from resources.rnn_utils import DatasetRNN


def plot_dynamics(
  choices: np.ndarray,
  rewards: np.ndarray,
  timeseries: Tuple[np.ndarray],
  timeseries_name: str,
  labels: Optional[Tuple[str]] = None,
  title: str = '',
  x_label = 'Trials',
  fig_ax = None,
  compare=False,
  color=None,
  x_axis_info=True,
  y_axis_info=True,
  reward_range=(0, 1),  # New parameter for reward range
  ):
  """Plot data from a single behavioral session of the bandit task.

  Args:
    choices: The choices made by the agent
    rewards: The rewards received by the agent (can be binary or float)
    timeseries: The dynamical value of interest on each arm
    timeseries_name: The name of the timeseries
    labels: The labels for the lines in the plot
    fig_ax: A tuple of a matplotlib figure and axis to plot on
    compare: If True, plot multiple timeseries on the same plot
    color: A list of colors to use for the plot; at least as long as the number of timeseries
    reward_range: Tuple specifying the min and max possible reward values
  """

  if color == None:
    color = [None]*len(timeseries)
  
  # Make the plot
  if fig_ax is None:
    fig, ax = plt.subplots(1, figsize=(10, 3))
  else:
    fig, ax = fig_ax
    
  if compare:
    if timeseries.ndim==2:
      timeseries = np.expand_dims(timeseries, -1)
    if timeseries.ndim!=3 or timeseries.shape[-1]!=1:
      raise ValueError('If compare: timeseries must be of shape (agent, timesteps, 1).')
  else:
    if timeseries.ndim!=2:
      raise ValueError('timeseries must be of shape (timesteps, n_actions).')
                       
  if not compare:
    # choices = np.expand_dims(choices, 0)
    timeseries = np.expand_dims(timeseries, 0)
  
  for i in range(timeseries.shape[0]):
    if labels is not None:
      if timeseries[i].ndim == 1:
        timeseries[i] = timeseries[i, :, None]
      if not compare:
        if len(labels) != timeseries[i].shape[1]:
          raise ValueError('labels length must match timeseries.shape[1].')
      else:
        if timeseries[i].shape[1] != 1:
          raise ValueError('If compare: timeseries.shape[1] must be 1.')
        if len(labels) != timeseries.shape[0]:
          raise ValueError('If compare: labels length must match timeseries.shape[0].')
      for ii in range(timeseries[i].shape[-1]):
          label = labels[ii] if not compare else labels[i]
          ax.plot(timeseries[i, :, ii], label=label, color=color[i])
      ax.legend(bbox_to_anchor=(1, 1))
    else:  # Skip legend.
      ax.plot(timeseries[i], color=color[i])
  
  # Plot ticks relating to whether the option was chosen (factual) or not (counterfactual) and reward value
  min_y, max_y = np.min(timeseries), np.max(timeseries)
  diff_min_max = np.max((1e-1, max_y - min_y))
  
  x = np.arange(len(choices))
  chosen_y = min_y - 1e-1  # Lower position for chosen (bigger tick)
  not_chosen_y = max_y + 1e-1  # Upper position for not chosen (smaller tick)

  # Set up color mapping for rewards
#   reward_min, reward_max = reward_range
  reward_min, reward_max = reward_range[0], reward_range[1]
  norm = Normalize(vmin=reward_min, vmax=reward_max)
  
  # Use RdYlGn colormap (Red-Yellow-Green) reversed so red is negative, green is positive
  cmap = cm.RdYlGn
  
  # Check if rewards are binary or continuous
  is_binary = reward_min == 0 and reward_max == 1 and np.all(np.isin(rewards, [-1, 0, 1]))
  
  if is_binary:
    # Handle binary rewards (backward compatibility)
    # Plot ticks for chosen options
    ax.scatter(x[(choices == 1) & (rewards == 1)], np.full(sum((choices == 1) & (rewards == 1)), chosen_y), 
               color='green', s=120, marker='|')  # Large green tick for chosen reward
    ax.scatter(x[(choices == 1) & (rewards == 0)], np.full(sum((choices == 1) & (rewards == 0)), chosen_y), 
               color='red', s=80, marker='|')  # Large red tick for chosen penalty

    # Plot ticks for not chosen options
    # ax.scatter(x[(choices == 0) & (rewards == 1)], np.full(sum((choices == 0) & (rewards == 1)), not_chosen_y), 
    #            color='green', s=120, marker='|')  # Small green tick
    # ax.scatter(x[(choices == 0) & (rewards == 0)], np.full(sum((choices == 0) & (rewards == 0)), not_chosen_y), 
    #            color='red', s=80, marker='|')  # Small red tick
  else:
    # Handle continuous rewards
    # Plot ticks for chosen options
    chosen_mask = (choices == 1)
    if np.any(chosen_mask):
      chosen_rewards = rewards[chosen_mask]
      chosen_colors = cmap(norm(chosen_rewards))
      # Size based on absolute reward value (larger for more extreme rewards)
      chosen_sizes = 60 + 60 * np.abs(chosen_rewards) / max(abs(reward_min), abs(reward_max))
      ax.scatter(x[chosen_mask], np.full(sum(chosen_mask), chosen_y), 
                 c=chosen_colors, s=chosen_sizes, marker='|', alpha=0.8)

    # Plot ticks for not chosen options
    # not_chosen_mask = (choices == 0)
    # if np.any(not_chosen_mask):
    #   not_chosen_rewards = rewards[not_chosen_mask]
    #   not_chosen_colors = cmap(norm(not_chosen_rewards))
    #   # Smaller size for not chosen options
    #   not_chosen_sizes = 40 + 40 * np.abs(not_chosen_rewards) / max(abs(reward_min), abs(reward_max))
    #   ax.scatter(x[not_chosen_mask], np.full(sum(not_chosen_mask), not_chosen_y), 
    #              c=not_chosen_colors, s=not_chosen_sizes, marker='|', alpha=0.6)
    
    # Add a colorbar to show the reward scale
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax, shrink=0.6)
    cbar.set_label('Reward Value')
    
  if x_axis_info:
    ax.set_xlabel(x_label)
  else:
    # ax.set_xticks(np.linspace(1, len(timeseries), 5))
    ax.set_xticklabels(['']*5)
    
  if y_axis_info:
    ax.set_ylabel(timeseries_name)
  else:
    ax.set_yticks(np.linspace(0, 1, 5))
    ax.set_yticklabels(['']*5)
    
  ax.set_title(title)
  
  return fig, ax


def plot_session(
    agents: Dict[str, Union[AgentSpice, AgentNetwork, AgentQ]], 
    experiment: Union[BanditSession, np.ndarray], 
    labels: List[str] = None, 
    save: str = None, 
    display_choice: int = 0,
    reward_range: List[float] = [0, 1],
    signals_to_plot: List[str] = ['x_value_reward', 'x_learing_rate_reward', 'x_value_choice'],
    ):    
    # plot the dynamcis associated with the first arm
    
    y_axis = True
    x_axis = True
    
    # valid keys in agent dictionary
    valid_keys_color_pairs = {'groundtruth': 'tab:blue', 'rnn': 'tab:orange', 'sindy': 'tab:pink', 'spice': 'tab:pink', 'benchmark':'tab:grey'}
    
    n_actions = agents[list(agents.keys())[0]]._n_actions
    if isinstance(experiment, BanditSession):
        choices = np.eye(n_actions)[experiment.choices.astype(int)][:, display_choice]
        rewards = experiment.rewards[:, display_choice]
    elif isinstance(experiment, np.ndarray) or isinstance(experiment, torch.Tensor):
        if isinstance(experiment, torch.Tensor):
            experiment = experiment.detach().cpu().numpy()
        assert experiment.ndim == 2, 'Experiment data should have only two dimensions -> (timesteps, features)'
        # choices = experiment[:, :n_actions].argmax(axis=-1)
        # rewards = np.array([exp[choices[i]] for i, exp in enumerate(experiment[:, n_actions:2*n_actions])])  
        choices = experiment[:, display_choice]
        rewards = experiment[:, n_actions]  
    
    list_probs = []
    list_q_value = []
    list_signals = {signal: [] for signal in signals_to_plot}
    values_signals = {signal: [] for signal in signals_to_plot}

    colors = []
    if labels is None:
        labels = []
        
    for valid_key in valid_keys_color_pairs:        
        # get q-values from agent
        if valid_key in [key.lower() for key in agents]:
            qs, probs, _ = get_update_dynamics(experiment, agents[valid_key], additional_signals=signals_to_plot)
            list_probs.append(np.expand_dims(probs, 0))
            list_q_value.append(np.expand_dims(qs[0], 0))
            for signal in signals_to_plot:
              list_signals[signal].append(np.expand_dims(qs[1][signal], 0))
            
            # get color from current agent
            colors.append(valid_keys_color_pairs[valid_key])
            
            if len(labels) < len(agents):
                # get labels from current agent
                labels.append(valid_key)

    # concatenate all choice probs and q-values
    probs = np.concatenate(list_probs, axis=0)
    q_value = np.concatenate(list_q_value, axis=0)
    for signal in signals_to_plot:
      values_signals[signal] = np.concatenate(list_signals[signal], axis=0)

    fig, axs = plt.subplots(2 + len(signals_to_plot), 1, figsize=(10, 10))
    axs_row = 0
    fig_col = None
    
    plot_dynamics(
        compare=True,
        choices=choices,
        rewards=rewards,
        timeseries=probs[:, :, display_choice],
        timeseries_name='$P(action)$',
        color=colors,
        labels=labels,
        fig_ax=(fig, axs[axs_row, fig_col]) if fig_col is not None else (fig, axs[axs_row]),
        x_axis_info=x_axis,
        y_axis_info=y_axis,
        reward_range=reward_range,
        )
    axs_row += 1
    
    plot_dynamics(
        compare=True,
        choices=choices,
        rewards=rewards,
        timeseries=q_value[:, :, display_choice],
        timeseries_name='$q$',
        color=colors,
        fig_ax=(fig, axs[axs_row, fig_col]) if fig_col is not None else (fig, axs[axs_row]),
        y_axis_info=x_axis,
        x_axis_info=y_axis,
        reward_range=reward_range,
        )
    axs_row += 1
    
    for signal in signals_to_plot:
      plot_dynamics(
          compare=True,
          choices=choices,
          rewards=rewards,
          timeseries=values_signals[signal][:, :, display_choice],
          timeseries_name=signal,
          color=colors,
          fig_ax=(fig, axs[axs_row, fig_col]) if fig_col is not None else (fig, axs[axs_row]),
          x_axis_info=x_axis,
          y_axis_info=y_axis,
          reward_range=reward_range,
          )
      axs_row += 1
    
    if save is not None:
        plt.savefig(save, dpi=300)
    
    return fig, axs


import numpy as np

    
def log_likelihood(data: np.ndarray, probs: np.ndarray, **kwargs):
    # data: array of binary observations (0 or 1)
    # probs: array of predicted probabilities for outcome 1 
    
    # Sum over all data points
    # return np.sum(np.sum(data * np.log(probs), axis=-1), axis=axis) / normalization
    # Ensure probabilities are within a valid range to prevent log(0)
    epsilon = 1e-9
    probs = np.clip(probs, epsilon, 1 - epsilon)
    
    # Calculate log-likelihood for each observation
    log_likelihoods = data * np.log(probs)# + (1 - data) * np.log(1 - probs)
    # log_likelihoods = data * np.log(probs)
    # log_likelihoods = np.sum(data * np.log(probs), axis=-1)
    
    # Sum log-likelihoods over all observations
    return np.sum(log_likelihoods)

def bayesian_information_criterion(data: np.ndarray, probs: np.ndarray, n_parameters: int, ll: np.ndarray = None, **kwargs):
    # data: array of binary observations (0 or 1)
    # probs: array of predicted probabilities for outcome 1
    # n_parameters: integer number of trainable model parameters
    
    if ll is None:
        ll = log_likelihood(data=data, probs=probs)
    
    n_samples = (data[:, 0] != -1).sum()
    return -2 * ll + n_parameters * np.log(n_samples)

def akaike_information_criterion(data: np.ndarray, probs: np.ndarray, n_parameters: int, ll: np.ndarray = None, **kwargs):
    # data: array of binary observations (0 or 1)
    # probs: array of predicted probabilities for outcome 1
    # n_parameters: integer number of trainable model parameters
    
    if ll is None:
        ll = log_likelihood(data=data, probs=probs)
    
    return -2 * ll + 2 * n_parameters

def get_scores(data: np.ndarray, probs: np.ndarray, n_parameters: int, **kwargs) -> float:
        ll = log_likelihood(data=data, probs=probs)
        bic = bayesian_information_criterion(data=data, probs=probs, n_parameters=n_parameters, ll=ll)
        aic = akaike_information_criterion(data=data, probs=probs, n_parameters=n_parameters, ll=ll)
        return -ll, aic, bic



import sys
import os

from typing import List

import pandas as pd
import numpy as np
import warnings
from copy import copy

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from resources.rnn_utils import DatasetRNN
from resources.bandits import BanditSession

def convert_dataset(
    file: str, 
    device = None, 
    sequence_length: int = None,
    df_participant_id: str = 'session',
    df_block: str = 'block',
    df_experiment_id: str = 'experiment',
    df_choice: str = 'choice',
    df_reward: str = 'reward',
    additional_inputs: List[str] = None,
    ):
    df = pd.read_csv(file, index_col=None)
    
    # replace all nan values with -1
    df = df.replace(np.nan, -1)
    
    original_df = copy(df)
    
    groupby_kw = [df_participant_id, df_experiment_id, df_block]
    
    # convert all participant ids to numeric values if necessary
    # if not pd.to_numeric(df[df_participant_id], errors='coerce').notna().all():
    # Map unique participant IDs to numeric values
    id_map = {pid: idx for idx, pid in enumerate(df[df_participant_id].unique())} #this line
    # Replace participant IDs with numeric values
    df[df_participant_id] = df[df_participant_id].map(id_map)    
    
    if not df_experiment_id in df.columns:
        df[df_experiment_id] = 0
    else:
        # convert all experiment ids to numeric values if necessary
        # if not all(pd.to_numeric(df[df_experiment_id], errors='coerce').notna().all()):
        # Map unique participant IDs to numeric values
        id_map = {eid: idx for idx, eid in enumerate(df[df_experiment_id].unique())}
        # Replace experiment IDs with numeric values
        df[df_experiment_id] = df[df_experiment_id].map(id_map)
    
    if not df_block in df.columns:
        df[df_block] = 0
        original_df[df_block] = 0
    else:
        # normalize block numbers
        blocks_max = df[df_block].max()
        blocks_min = df[df_block].min()
        df[df_block] = (df[df_block] - blocks_min) / (blocks_max - blocks_min)
        
    # get maximum number of trials per participant
    n_groups = len(df.groupby(groupby_kw).size())
    max_trials = df.groupby(groupby_kw).size().max()
    
    # let actions begin from 0
    choices = df[df_choice].values
    choice_min = np.nanmin(choices[choices != -1])
    choices[choices != -1] = choices[choices != -1] - choice_min
    df[df_choice] = choices
    original_df[df_choice] = choices
    
    # number of possible actions
    n_actions = int(df[df_choice].max() + 1)
    
    # get all columns with rewards
    if df_reward + '_0' in df.columns:
        # counterfactual dataset
        reward_cols = []
        for reward_column in df.columns:
            if df_reward in reward_column:
                reward_cols.append(reward_column)
    else:
        reward_cols = [df_reward]
    
    # normalize rewards
    r_min, r_max = [], []
    for reward_column in reward_cols:
        r_min.append(df[reward_column].min())
        r_max.append(df[reward_column].max())
    r_min = np.min(r_min)
    r_max = np.max(r_max)
    for reward_column in reward_cols:
        df[reward_column] = (df[reward_column] - r_min) / (r_max - r_min)
    
    # check whether the additional inputs are in the dataset
    additional_inputs_xs = 0
    additional_input_not_found = []
    if isinstance(additional_inputs, list) or isinstance(additional_inputs, tuple):
        for additional_input in additional_inputs:
            if additional_input in df.columns:
                additional_inputs_xs += 1
            else:
                additional_input_not_found.append(additional_input)

    if len(additional_input_not_found) > 0:
        warnings.warn(f"Some additional input signals ({additional_input_not_found}) were not found in the dataset. These signals will be ignored.")
        for not_found in additional_input_not_found:
            additional_inputs.remove(not_found)
    
    # convert non-numerical items in additional inputs
    if additional_inputs:
        for index, additional in enumerate(additional_inputs):
            if isinstance(df[additional].values[0], str) and not df[additional].values[0].isdigit():
                # Map unique participant IDs to numeric values
                id_map = {pid: idx for idx, pid in enumerate(df[additional].unique())} #this line
                # Replace participant IDs with numeric values
                df[additional] = df[additional].map(id_map)    
                
                print(f"Values in column {additional} in dataset {file} are not numerical. Converted values are:")
                print(id_map)
            
    n_ids = 0
    if df_participant_id is not None:
        n_ids += 1
    if df_experiment_id is not None:
        n_ids += 1
        
    xs = np.zeros((n_groups, max_trials, n_actions*2 + n_ids + additional_inputs_xs + 1 if df_block is not None else 0)) - 1
    ys = np.zeros((n_groups, max_trials, n_actions)) - 1
    
    probs_choice = np.zeros((n_groups, max_trials, n_actions)) - 1
    values_action = np.zeros((n_groups, max_trials, n_actions)) - 1
    values_reward = np.zeros((n_groups, max_trials, n_actions)) - 1
    values_choice = np.zeros((n_groups, max_trials, n_actions)) - 1
    
    experiment_list = []
    # for index_group, participant_id in enumerate(participant_ids):
    for index_group, group in enumerate(df.groupby(groupby_kw)):
        
        choice = np.eye(n_actions)[group[1][df_choice].values.astype(int)]
        rewards = np.zeros((len(choice), n_actions)) - 1
        for index_column, reward_column in enumerate(reward_cols):
            # add reward into the correct column
            reward = group[1][reward_column].values
            if len(reward_cols) > 1:
                # counterfactual data
                rewards[:, index_column] = reward
            else:
                for index_choice, ch in enumerate(choice):
                    rewards[index_choice, np.argmax(ch)] = reward[index_choice]

        # write arrays for DatasetRNN
        xs[index_group, :len(choice), :n_actions] = choice
        xs[index_group, :len(choice), n_actions:n_actions*2] = rewards
        
        # add all grouping variables to the end
        xs[index_group, :, -3] = group[1][df_block].values[0]
        xs[index_group, :, -2] = group[1][df_experiment_id].values[0]
        xs[index_group, :, -1] = group[1][df_participant_id].values[0]
        ys[index_group, :len(choice)-1] = choice[1:]
        
        # write additional inputs after choices and rewards and before grouping variables
        if additional_inputs is not None and len(additional_inputs) > 0:
            for index, additional_input in enumerate(additional_inputs):
                xs[index_group, :len(choice), n_actions*2+index] = group[1][additional_input].values
        
        experiment = BanditSession(
            choices=group[1][df_choice].values,
            rewards=rewards,
            session=np.full((*rewards.shape[:-1], 1), index_group),
            # reward_probabilities=np.zeros_like(choice)+0.5,
            # q=np.zeros_like(choice)+0.5,
            n_trials=len(choice)
        )
        
        # get update dynamics if available - only for generated data with e.g. utils/create_dataset.py
        if 'choice_prob_0' in df.columns:
            for index_choice in range(n_actions):
                probs_choice[index_group, :len(choice), index_choice] = group[1][f'choice_prob_{index_choice}'].values
                values_action[index_group, :len(choice), index_choice] = group[1][f'action_value_{index_choice}'].values
                values_reward[index_group, :len(choice), index_choice] = group[1][f'reward_value_{index_choice}'].values
                values_choice[index_group, :len(choice), index_choice] = group[1][f'choice_value_{index_choice}'].values
        
        experiment_list.append(experiment)
        
    return DatasetRNN(xs, ys, device=device, sequence_length=sequence_length), experiment_list, original_df, (probs_choice, values_action, values_reward, values_choice)


from typing import NamedTuple, Union, Optional, Dict, Callable, Tuple, List

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from copy import copy, deepcopy
import torch
import pickle
import dill
from tqdm import tqdm
import pysindy as ps

from .rnn import BaseRNN, ParameterModule
from .rnn_utils import DatasetRNN

# Setup so that plots will look nice
small = 15
medium = 18
large = 20
plt.rc('axes', titlesize=large)
plt.rc('axes', labelsize=medium)
plt.rc('xtick', labelsize=small)
plt.rc('ytick', labelsize=small)
plt.rc('legend', fontsize=small)
plt.rc('figure', titlesize=large)
mpl.rcParams['grid.color'] = 'none'
mpl.rcParams['axes.facecolor'] = 'white'
plt.rcParams['svg.fonttype'] = 'none'

###################################
# CONVENIENCE FUNCTIONS.          #
###################################


def check_in_0_1_range(x, name):
  if not (0 <= x <= 1):
    raise ValueError(
        f'Value of {name} must be in [0, 1] range. Found value of {x}.')


###################################
# GENERATIVE FUNCTIONS FOR AGENTS #
###################################

class Agent:
  """An agent that runs simple Q-learning for the two-armed bandit task.

  Attributes:
    alpha: The agent's learning rate
    beta: The agent's softmax temperature
    q: The agent's current estimate of the reward probability on each arm
  """

  def __init__(
      self,
      n_actions: int = 2,
      beta_reward: float = 3.,
      alpha_reward: float = 0.2,
      parameter_variance: Union[Dict[str, float], float] = 0.,
      deterministic: bool = True,
      ):
    """Update the agent after one step of the task.
    
    Args:
      alpha_reward (float): Baseline learning rate between 0 and 1.
      beta_reward (float): softmax inverse noise temperature. Regulates the noise in the decision-selection.
      n_actions: number of actions (default=2)
      parameter_variance (float): sets a variance around the model parameters' mean values to sample from a normal distribution e.g. at each new session. 0: no variance, -1: std = mean
    """
    
    self._list_params = ['beta_reward', 'alpha_reward']
    
    self._mean_beta_reward = beta_reward
    self._mean_alpha_reward = alpha_reward
    
    self._alpha_reward = alpha_reward
    
    self._betas = {}
    self._betas['x_value_reward'] = beta_reward
    
    self._n_actions = n_actions
    self._q_init = 0.5
    self._parameter_variance = self.check_parameter_variance(parameter_variance)
    self._deterministic = deterministic
    
    check_in_0_1_range(alpha_reward, 'alpha')
    
  def check_parameter_variance(self, parameter_variance):
    if isinstance(parameter_variance, float):
      par_var_dict = {}
      for key in self._list_params:
        par_var_dict[key] = parameter_variance
      parameter_variance = par_var_dict
    elif isinstance(parameter_variance, dict):
      # check that all keys in parameter_variance are valid
      not_valid_keys = []
      for key in parameter_variance:
        if not key in self._list_params:
          not_valid_keys.append(key)
      if len(not_valid_keys) > 0:
        raise ValueError(f'Some keys in parameter_variance are not valid ({not_valid_keys}). Valid keys are {self._list_params}')
      # check that all parameters are available - set to 0 if a parameter is not available
      for key in self._list_params:
        if not key in parameter_variance:
          parameter_variance[key] = 0.
    return parameter_variance
  
  def new_sess(self, **kwargs):
    """Reset the agent for the beginning of a new session."""
    
    self._state = {
      'x_value_reward': np.full((self._n_actions, 1), self._q_init),
      'x_value_choice': np.zeros((self._n_actions, 1)),
      'x_learning_rate_reward': np.zeros((self._n_actions, 1)),
    }
    
  def get_choice_probs(self) -> np.ndarray:
    """Compute the choice probabilities as softmax over q."""
    decision_variable = np.exp(self.q)
    choice_probs = decision_variable / np.sum(decision_variable)
    return choice_probs.reshape(self._n_actions)

  def get_choice(self):
    """Sample choice."""
    choice_probs = self.get_choice_probs()
    if self._deterministic:
      return np.argmax(choice_probs)
    else:
      return np.random.choice(self._n_actions, p=choice_probs)

  def update(self, choice: int, reward: np.ndarray, *args, **kwargs):
    """Update the agent after one step of the task.

    Args:
      choice: The choice made by the agent. 0 or 1
      reward: The reward received by the agent. 0 or 1
    """
    
    # adjust learning rates for every received reward
    alpha = np.zeros_like(self._state['x_learning_rate_reward'])
    rpe = np.zeros_like(self._state['x_learning_rate_reward'])
    for action in range(self._n_actions):
      if action == choice:
        current_reward = reward[action] if np.min(reward) > -1 else reward[choice]
        alpha[action] = self._alpha_reward
        
        # Reward-prediction-error
        rpe[action] = current_reward - self._state['x_value_reward'][action]
      
    # (Counterfactual) Reward update
    reward_update = alpha * rpe
    
    # Update memory state
    self._state['x_value_reward'] += reward_update
    self._state['x_learning_rate_reward'] = alpha
  
  def get_state_value(self, state: str, multiply_with_beta: bool = True):
    if multiply_with_beta and state in self._betas:
      return (self._state[state] * self._betas[state]).reshape(self._n_actions)
    else:
      return self._state[state].reshape(self._n_actions)

  @property
  def q(self):
    return (self._state['x_value_reward']*self._betas['x_value_reward']).reshape(self._n_actions)


class AgentQ(Agent):
  """An agent that runs simple Q-learning for the y-maze tasks.

  Attributes:
    alpha: The agent's learning rate
    beta: The agent's softmax temperature
    q: The agent's current estimate of the reward probability on each arm
  """

  def __init__(
      self,
      n_actions: int = 2,
      beta_reward: float = 3.,
      alpha_reward: float = 0.2,
      alpha_penalty: float = -1,
      beta_counterfactual: float = 0.,
      alpha_counterfactual_reward: float = 0.,
      alpha_counterfactual_penalty: float = 0.,
      beta_choice: float = 0.,
      alpha_choice: float = 1.,
      forget_rate: float = 0.,
      parameter_variance: Union[Dict[str, float], float] = 0.,
      deterministic: bool = False,
      ):
    """Update the agent after one step of the task.
    
    Args:
      alpha (float): Baseline learning rate between 0 and 1.
      beta (float): softmax inverse noise temperature. Regulates the noise in the decision-selection.
      n_actions: number of actions (default=2)
      forget_rate (float): rate at which q values decay toward the initial values (default=0)
      perseverance_bias (float): rate at which q values move toward previous action (default=0)
      alpha_penalty (float): separate learning rate for negative outcomes
      parameter_variance (float): sets a variance around the model parameters' mean values to sample from a normal distribution e.g. at each new session. 0: no variance, -1: std = mean
    """
    
    super().__init__(n_actions=n_actions, parameter_variance=parameter_variance, deterministic=deterministic)
    
    self._list_params = ['beta_reward', 'alpha_reward', 'alpha_penalty', 'alpha_counterfactual', 'beta_choice', 'alpha_choice', 'forget_rate']
    
    self._mean_beta_reward = beta_reward
    self._mean_alpha_reward = alpha_reward
    self._mean_alpha_penalty = alpha_penalty if alpha_penalty >= 0 else alpha_reward
    self._mean_forget_rate = forget_rate
    self._mean_beta_choice = beta_choice
    self._mean_alpha_choice = alpha_choice
    self._mean_alpha_counterfactual_reward = alpha_counterfactual_reward
    self._mean_alpha_counterfactual_penalty = alpha_counterfactual_penalty
    
    self._alpha_reward = alpha_reward
    self._alpha_penalty = alpha_penalty if alpha_penalty >= 0 else alpha_reward
    self._forget_rate = forget_rate
    self._alpha_choice = alpha_choice
    self._alpha_counterfactual_reward = alpha_counterfactual_reward
    self._alpha_counterfactual_penalty = alpha_counterfactual_penalty
    
    self._betas['x_value_reward'] = beta_reward
    self._betas['x_value_choice'] = beta_choice
    
    self._n_actions = n_actions
    self._parameter_variance = self.check_parameter_variance(parameter_variance)
    self._q_init = 0.5
    
    self.new_sess()

    check_in_0_1_range(alpha_reward, 'alpha')
    if alpha_penalty >= 0:
      check_in_0_1_range(alpha_penalty, 'alpha_penalty')
    check_in_0_1_range(alpha_counterfactual_reward, 'alpha_countefactual_reward')
    check_in_0_1_range(alpha_counterfactual_penalty, 'alpha_countefactual_penalty')
    check_in_0_1_range(alpha_choice, 'alpha_choice')
    check_in_0_1_range(forget_rate, 'forget_rate')
      
  def new_sess(self, sample_parameters=False, **kwargs):
    """Reset the agent for the beginning of a new session."""
    # self._q = np.full(self._n_actions, self._q_init)
    # self._c = np.zeros(self._n_actions)
    # self._alpha = np.zeros(self._n_actions)
    
    self._state = {
      'x_value_reward': np.full(self._n_actions, self._q_init),
      'x_value_choice': np.zeros(self._n_actions),
      'x_learning_rate_reward': np.zeros(self._n_actions),
    }
    
    # sample new parameters
    if sample_parameters:
      sanity = False
      while not sanity:
        # sample new parameters until all sanity checks are passed
        self._betas['x_value_reward'] = np.clip(np.random.normal(self._mean_beta_reward, self._mean_beta_reward/2 if self._parameter_variance['beta_reward'] == -1 else self._parameter_variance['beta_reward']), 0, 2*self._mean_beta_reward)
        self._betas['x_value_choice'] = np.clip(np.random.normal(self._mean_beta_choice, self._mean_beta_choice/2 if self._parameter_variance['beta_choice'] == -1 else self._parameter_variance['beta_choice']), 0, 2*self._mean_beta_choice)
        self._alpha_reward = np.clip(np.random.normal(self._mean_alpha_reward, self._mean_alpha_reward/2 if self._parameter_variance['alpha_reward'] == -1 else self._parameter_variance['alpha_reward']), 0 , 1)
        self._alpha_penalty = np.clip(np.random.normal(self._mean_alpha_penalty, self._mean_alpha_penalty/2 if self._parameter_variance['alpha_penalty'] == -1 else self._parameter_variance['alpha_penalty']), 0, 1)
        self._alpha_choice = np.clip(np.random.normal(self._mean_alpha_choice, self._mean_alpha_choice/2 if self._parameter_variance['alpha_choice'] == -1 else self._parameter_variance['alpha_choice']), 0, 1)
        self._alpha_counterfactual = np.clip(np.random.normal(self._mean_alpha_counterfactual, self._mean_alpha_counterfactual/2 if self._parameter_variance['alpha_counterfactual'] == -1 else self._parameter_variance['alpha_counterfactual']), 0, 1)
        self._forget_rate = np.clip(np.random.normal(self._mean_forget_rate, self._mean_forget_rate/2 if self._parameter_variance['forget_rate'] == -1 else self._parameter_variance['forget_rate']), 0, 1)

        # sanity checks
        max_learning_rate = self._alpha_reward <= 1 and self._alpha_penalty <= 1
        min_learning_rate = self._alpha_reward >= 0 and self._alpha_penalty >= 0 
        sanity = max_learning_rate and min_learning_rate

  def update(self, choice: int, reward: np.ndarray, *args, **kwargs):
    """Update the agent after one step of the task.

    Args:
      choice: The choice made by the agent. 0 or 1
      reward: The reward received by the agent. 0 or 1
    """
    
    # Reward-based updates
    non_chosen_action = np.arange(self._n_actions) != choice
    
    # adjust learning rates for every received reward
    alpha = np.zeros_like(reward)
    rpe = np.zeros_like(reward)
    for action in range(self._n_actions):
      if action == choice:
        # factual case
        alpha_r = self._alpha_reward
        alpha_p = self._alpha_penalty
      else: 
        # counterfactual case
        # counterfactual learning rate applies in case of foregone reward
        # no counterfactual learning in case of foregone penalty
        alpha_r = self._alpha_counterfactual_reward
        alpha_p = self._alpha_counterfactual_penalty
        
      # asymmetric learning rates
      current_reward = reward[action] if np.min(reward) > -1 else reward[choice]
      alpha[action] = alpha_r if current_reward > 0.5 else alpha_p
      
      # Reward-prediction-error
      r = current_reward if action==choice else 1-current_reward
      rpe[action] = r - self._state['x_value_reward'][action]
      
    # (Counterfactual) Reward update
    reward_update = alpha * rpe
    
    # Forgetting - restore q-values of non-chosen actions towards the initial value
    forget_update = self._forget_rate * (self._q_init - self._state['x_value_reward'][non_chosen_action])

    # Choice-Perseverance: Action-based updates
    cpe = np.eye(self._n_actions)[choice] - self._state['x_value_choice']
    
    # Update memory state
    self._state['x_value_reward'][non_chosen_action] += reward_update[non_chosen_action] + forget_update
    self._state['x_value_reward'][choice] += reward_update[choice]
    self._state['x_value_choice'] += self._alpha_choice * cpe
    self._state['x_learning_rate_reward'] = alpha

  @property
  def q(self):
    return (self._state['x_value_reward']*self._betas['x_value_reward'] + self._state['x_value_choice']*self._betas['x_value_choice']).reshape(self._n_actions)


class AgentQ_SampleZeros(AgentQ):
  """An agent that runs simple Q-learning for the y-maze tasks.

  Attributes:
    alpha: The agent's learning rate
    beta: The agent's softmax temperature
    q: The agent's current estimate of the reward probability on each arm
  """

  def __init__(
      self,
      n_actions: int = 2,
      beta_reward: float = 3.,
      alpha_reward: float = 0.2,
      alpha_penalty: float = -1,
      alpha_counterfactual: float = 0.,
      beta_choice: float = 0.,
      alpha_choice: float = 1.,
      forget_rate: float = 0.,
      parameter_variance: Union[Dict[str, float], float] = 0.,
      beta_distribution: np.ndarray = (0.7, 1.0),
      zero_threshold: float = 0.1,
      ):
    
    super(AgentQ_SampleZeros, self).__init__(
      n_actions=n_actions,
      beta_reward=beta_reward,
      alpha_reward=alpha_reward,
      alpha_penalty=alpha_penalty,
      alpha_counterfactual=alpha_counterfactual,
      beta_choice=beta_choice,
      alpha_choice=alpha_choice,
      forget_rate=forget_rate,
      parameter_variance=parameter_variance,
      )
    
    self._beta_distribution = beta_distribution
    self._zero_threshold = zero_threshold 
  
  def new_sess(self, sample_parameters=False, **kwargs):
    """Reset the agent for the beginning of a new session."""
    
    super(AgentQ_SampleZeros, self).new_sess()
    
    # sample new parameters
    if sample_parameters:
      # sample scaling parameters (inverse noise temperatures)
      self._betas['x_value_reward'], self._betas['x_value_choice'] = 0, 0
      while self._betas['x_value_reward'] <= self._zero_threshold and self._betas['x_value_choice'] <=  self._zero_threshold:
        self._betas['x_value_reward'] = np.random.rand()
        self._betas['x_value_choice'] = np.random.rand()
        # apply zero-threshold if applicable
        self._betas['x_value_reward'] = self._betas['x_value_reward'] * 2 * self._mean_beta_reward if self._betas['x_value_reward'] > self._zero_threshold else 0
        self._betas['x_value_choice'] = self._betas['x_value_choice'] * 2 * self._mean_beta_choice if self._betas['x_value_choice'] > self._zero_threshold else 0
      
      # sample auxiliary parameters
      self._forget_rate = np.random.rand()
      self._forget_rate = self._forget_rate * (self._forget_rate > self._zero_threshold)
      
      # sample learning rate; don't zero out; only check for applicability of asymmetric learning rates
      self._alpha_reward = np.random.rand()
      self._alpha_penalty = np.random.rand()
      # set alpha_reward = alpha_penalty if (alpha_reward - alpha_penalty) < zero_threshold
      if np.abs(self._alpha_reward - self._alpha_penalty) < self._zero_threshold:
        alpha_mean = np.mean((self._alpha_reward, self._alpha_penalty))
        self._alpha_reward = alpha_mean
        self._alpha_penalty = alpha_mean


class AgentNetwork(Agent):
  """A class that allows running a pretrained RNN as an agent.

  Attributes:
      model: A PyTorch module representing the RNN architecture
  """

  def __init__(
    self,
    model_rnn: BaseRNN,
    n_actions: int = 2,
    device = torch.device('cpu'),
    deterministic: bool = True,
    ):
      """Initialize the agent network.

      Args:
          model: A PyTorch module representing the RNN architecture
          n_actions: number of permitted actions (default = 2)
      """
      
      super().__init__(n_actions=n_actions)

      self._deterministic = deterministic

      self._model = model_rnn
      if model_rnn is not None:
        self._model = self._model.to(device)
        self._model.eval()

  def new_sess(self, participant_id: int = 0, experiment_id: int = 0, additional_embedding_inputs: np.ndarray = torch.zeros(0), **kwargs):
    """Reset the network for the beginning of a new session."""
    if not isinstance(participant_id, torch.Tensor):
      participant_id = torch.tensor(participant_id, dtype=int, device=self._model.device)[None]
    
    self._model.set_initial_state(batch_size=1)
    
    self._meta_data = torch.zeros((1, 2))
    self._meta_data[0, -1] = participant_id
    self._meta_data[0, -2] = experiment_id
    
    self._additional_meta_data = additional_embedding_inputs if isinstance(additional_embedding_inputs, torch.Tensor) else torch.tensor(additional_embedding_inputs, dtype=torch.float32)
    self._additional_meta_data = self._additional_meta_data.view(1, -1)
    
    self.set_state()

  def get_logit(self):
    """Return the value of the agent's current state."""
    betas = self.get_betas()
    if betas:
      logits = np.sum(
        np.concatenate([
          self._state[key] * betas[key] for key in self._state if key in betas 
          ]), 
        axis=0)
    else:
      logits = np.sum(
        np.concatenate([
          self._state[key] for key in self._state if 'x_value' in key
          ]),
        axis=0)
    return logits

  def update(self, choice: float, reward: float, block: int = 0, additional_inputs: np.ndarray = torch.zeros(0), **kwargs):
    choice = torch.eye(self._n_actions)[int(choice)]
    xs = torch.concat([choice, torch.tensor(reward), torch.tensor(additional_inputs), torch.tensor(block).view(1), self._meta_data.view(-1)]).view(1, 1, -1).to(device=self._model.device)
    with torch.no_grad():
      self._model(xs, self._model.get_state(detach=True))
    self.set_state()
  
  def set_state(self):
    self._state = self._model.get_state()
    self._betas = self.get_betas()
    
  def get_betas(self):
    if hasattr(self._model, 'betas'):
      betas = {}
      if hasattr(self._model, 'participant_embedding'):
        participant_embedding = self._model.participant_embedding(self._meta_data[..., -1].int().to(device=self._model.device).view(1, 1))
      for key in self._model.betas:
        betas[key] = self._model.betas[key]().item() if isinstance(self._model.betas[key], ParameterModule) else self._model.betas[key](participant_embedding).item()
      return betas
    return None

  def get_participant_ids(self):
    if hasattr(self._model, 'participant_embedding'):
      return tuple(np.arange(self._model.participant_embedding.num_embeddings).tolist())
  
  @property
  def q(self):
    return self.get_logit()
  
  
class AgentSpice(AgentNetwork):
  
  def __init__(
    self,
    model_rnn: BaseRNN,
    sindy_modules: Dict,
    n_actions: int,
    deterministic: bool = True,
  ):
    
    super().__init__(model_rnn=deepcopy(model_rnn), n_actions=n_actions, deterministic=deterministic)
    
    self._model.integrate_sindy(sindy_modules)
  
  def get_modules(self) -> Dict[str, Dict[int, ps.SINDy]]:
    return self._model.submodules_sindy
  
  def count_parameters(self, mapping_modules_values: dict = None) -> Dict[int, int]:
    """Count the number of non-zero parameters in each module for each participant. 
    Considers also beta values (if mapping_modules_values is given).
    
    Args:
        mapping_modules_values (dict, optional): Defines which module maps onto which value in the memory state (will be deprecated in newer versions because this information will be stored as an attribute in the RNN) 

    Returns:
        Dict[int, int]: Dictionary which maps the participant ID onto the respective number of parameters
    """
    
    submodules = self.get_modules()
    keys_submodules = list(submodules.keys())
    participant_ids = list(submodules[keys_submodules[0]].keys())
    n_parameters = {participant_id: 0 for participant_id in participant_ids}
    for participant_id in participant_ids:
      self.new_sess(participant_id=participant_id, additional_embedding_inputs=self._additional_meta_data)
      betas = self.get_betas()
      # count all non-zero coefficients in SINDy modules with considering the corresponding beta value which potentially can set all influences of this module to 0 
      for submodule in submodules:
        parameters_module = submodules[submodule][participant_id].coefficients()
        # n_parameters_module = n_parameters_module * (n_parameters_module > 0.05)
        if betas:
          #beta_value_module = betas[mapping_modules_values[submodule]]
          # n_parameters[participant_id] += (parameters_module * beta_value_module != 0).sum()
          n_parameters[participant_id] += (parameters_module != 0).sum()
        else:
          n_parameters[participant_id] += (parameters_module != 0).sum()
      if betas:
        # include beta parameters if non-zero
        for value in betas:
          if np.abs(betas[value]) > 1e-2:
            n_parameters[participant_id] += 1
    return n_parameters
  
  def get_participant_ids(self):
    modules = self.get_modules()
    return list(modules[list(modules.keys())[0]].keys())  
  
  def print_model(self, participant_id: int):
    for module in self.get_modules():
      self._model.submodules_sindy[module][participant_id].print()


################
# ENVIRONMENTS #
################


class Bandits:
  
  def __init__(self, *args, **kwargs):
    pass
  
  def step(self, choice):
    pass
  
  def new_sess(self):
    pass


class BanditsFlip(Bandits):
  """Env for 2-armed bandit task with reward probs that flip in blocks."""

  def __init__(
      self,
      block_flip_prob: float = 0.02,
      reward_prob_high: float = 0.8,
      reward_prob_low: float = 0.2,
      counterfactual: bool = False,
      **kwargs,
  ):
    
    super(BanditsFlip, self).__init__()
    
    # Assign the input parameters as properties
    self._block_flip_prob = block_flip_prob
    self._reward_prob_high = reward_prob_high
    self._reward_prob_low = reward_prob_low
    self._counterfactual = counterfactual
    # Choose a random block to start in
    self._block = np.random.binomial(1, 0.5)
    # Set up the new block
    self.new_block()

  def new_block(self):
    """Flip the reward probabilities for a new block."""
    # Flip the block
    self._block = 1 - self._block
    # Set the reward probabilites
    if self._block == 1:
      self.reward_probs = [self._reward_prob_high, self._reward_prob_low]
    else:
      self.reward_probs = [self._reward_prob_low, self._reward_prob_high]

  def step(self, choice: int = None):
    """Step the model forward given chosen action."""

    # Sample a reward with this probability
    reward = np.array([float(np.random.binomial(1, prob)) for prob in self.reward_probs], dtype=float)
    
    # Check whether to flip the block
    if np.random.binomial(1, self._block_flip_prob):
      self.new_block()

    # Return the reward
    choice_onehot = np.eye(self.n_actions)[choice]
    if self._counterfactual:
      return reward
    else:
      return choice_onehot * reward[choice] + (1-choice_onehot)*-1

  @property
  def n_actions(self) -> int:
    return 2
  

class BanditsSwitch(Bandits):
  """Env for 2-armed bandit task with fixed sets of reward probs that switch in blocks."""

  def __init__(
      self,
      block_flip_prob: float = 0.02,
      reward_prob_high: float = 0.75,
      reward_prob_low: float = 0.25,
      reward_prob_middle: float = 0.5,
      counterfactual: bool = False,
      **kwargs,
  ):
    
    super(BanditsSwitch, self).__init__()
    
    # Assign the input parameters as properties
    self._block_flip_prob = block_flip_prob
    self._reward_prob_high = reward_prob_high
    self._reward_prob_low = reward_prob_low
    self._reward_prob_middle = reward_prob_middle
    self._counterfactual = counterfactual
    self._n_blocks = 7
    
    # Choose a random block to start in
    self._block = np.random.randint(self._n_blocks)
    
    # Set up the new block
    self.new_block()

  def new_block(self):
    """Switch the reward probabilities for a new block."""
    
    # Choose a new random block
    block = np.random.randint(0, self._n_blocks)
    while block == self._block:
      block = np.random.randint(0, self._n_blocks)
    self._block = block
    
    # Set the reward probabilites
    if self._block == 0:
      self._reward_probs = [self._reward_prob_high, self._reward_prob_low]
    elif self._block == 1:
      self._reward_probs = [self._reward_prob_middle, self._reward_prob_middle]
    elif self._block == 2:
      self._reward_probs = [self._reward_prob_low, self._reward_prob_high]
    elif self._block == 3:
      self._reward_probs = [self._reward_prob_low, self._reward_prob_middle]
    elif self._block == 4:
      self._reward_probs = [self._reward_prob_middle, self._reward_prob_high]
    elif self._block == 5:
      self._reward_probs = [self._reward_prob_middle, self._reward_prob_low]
    elif self._block == 6:
      self._reward_probs = [self._reward_prob_high, self._reward_prob_middle]
      
  def step(self, choice: int = None):
    """Step the model forward given chosen action."""

    # Sample a reward with this probability
    reward = np.array([float(np.random.binomial(1, prob)) for prob in self.reward_probs], dtype=float)

    # Check whether to flip the block
    if np.random.uniform() < self._block_flip_prob:
      self.new_block()

    # Return the reward
    choice_onehot = np.eye(self.n_actions)[choice]
    if self._counterfactual:
      return reward
    else:
      return choice_onehot * reward[choice] + (1-choice_onehot)*-1

  @property
  def reward_probs(self) -> np.ndarray:
    return self._reward_probs.copy()

  @property
  def n_actions(self) -> int:
    return 2


class BanditsDrift(Bandits):
  """Environment for a drifting two-armed bandit task.

  Reward probabilities on each arm are sampled randomly between 0 and 1. On each
  trial, gaussian random noise is added to each.

  Attributes:
    sigma: A float, between 0 and 1, giving the magnitude of the drift
    reward_probs: Probability of reward associated with each action
    n_actions: number of actions available
  """

  def __init__(
      self,
      sigma: float,
      n_actions: int = 2,
      counterfactual: bool = False,
      **kwargs,
      ):
    """Initialize the environment."""
    
    super(BanditsDrift, self).__init__()
    
    # Check inputs
    if sigma < 0:
      msg = f'Argument sigma but must be greater than 0. Found: {sigma}.'
      raise ValueError(msg)

    # Initialize persistent properties
    self._sigma = sigma
    self._n_actions = n_actions
    self._counterfactual = counterfactual

    # Sample new reward probabilities
    self.new_sess()

  def new_sess(self):
    # Pick new reward probabilities.
    # Sample randomly between 0 and 1
    self._reward_probs = np.random.rand(self._n_actions)

  def step(self, choice: int) -> np.ndarray:
    """Run a single trial of the task.

    Args:
      choice: integer specifying choice made by the agent (must be less than
        n_actions.)

    Returns:
      reward: The reward to be given to the agent. 0 or 1.

    """

    # Sample reward with the probability of the chosen side
    reward = np.array([float(np.random.rand() < self._reward_probs[i]) for i in range(self._n_actions)])
    
    # Add gaussian noise to reward probabilities
    drift = np.random.normal(loc=0, scale=self._sigma, size=self._n_actions)
    self._reward_probs += drift
    
    # Fix reward probs that've drifted below 0 or above 1
    self._reward_probs = np.clip(self._reward_probs, 0, 1)

    # Return the reward
    choice_onehot = np.eye(self.n_actions)[choice]
    if self._counterfactual:
      return reward
    else:
      return choice_onehot * reward[choice] + (1-choice_onehot)*-1

  @property
  def reward_probs(self) -> np.ndarray:
    return self._reward_probs.copy()

  @property
  def n_actions(self) -> int:
    return self._n_actions


class BanditsFlip_eckstein2022(Bandits):
  """Env for 2-armed bandit task with reward probs that flip in blocks as used in Eckstein et al (2022).
  Additional flipping criteria: accumulated reward > threshold with threshold ~ Uniform(7,15)."""

  def __init__(
      self,
      reward_prob_high: float = 0.75,
      reward_prob_low: float = 0.,
      counterfactual: bool = False,
      **kwargs
  ):
    
    super(BanditsFlip_eckstein2022, self).__init__()
    
    # Assign the input parameters as properties
    self._block_flip_criteria = np.random.uniform(7, 15)
    self._accumulated_reward_block = 0
    self._accumulated_steps_block = 0
    
    self._reward_prob_high = reward_prob_high
    self._reward_prob_low = reward_prob_low
    
    self._counterfactual = counterfactual
    
    # Choose a random block to start in
    self._block = np.random.binomial(1, 0.5)
    
    # Set up the new block
    self.new_block()

  def new_sess(self):
    # Choose a random block to start in
    self._block = np.random.binomial(1, 0.5)
    self.new_block()

  def new_block(self):
    """Flip the reward probabilities for a new block."""
    # Flip the block
    self._block = 1 - self._block
    # Set the reward probabilites
    if self._block == 1:
      self.reward_probs = [self._reward_prob_high, self._reward_prob_low]
    else:
      self.reward_probs = [self._reward_prob_low, self._reward_prob_high]
      
    self._block_flip_criteria = np.random.uniform(7, 15)
    self._accumulated_reward_block = 0
    self._accumulated_steps_block = 0

  def step(self, choice: int = None):
    """Step the model forward given chosen action."""

    if self._accumulated_steps_block == 0 and choice == np.argmax(self.reward_probs):
      # first trial in a new block -> if the choice is correct -> always reward 
      reward = np.zeros(self.n_actions)
      reward[choice] = 1
    else:
      # Sample a reward with this probability
      reward = np.array([float(np.random.binomial(1, prob)) for prob in self.reward_probs], dtype=float)
    
    # Check whether to flip the block
    self._accumulated_reward_block += reward[choice]
    self._accumulated_steps_block += 1
    if self._accumulated_reward_block >= self._block_flip_criteria:
      self.new_block()
    
    # Return the reward
    choice_onehot = np.eye(self.n_actions)[choice]
    if self._counterfactual:
      return reward
    else:
      return choice_onehot * reward[choice] + (1-choice_onehot)*-1

  @property
  def n_actions(self) -> int:
    return 2
  

class Bandits_Standard(Bandits):
  """Env for 2-armed bandit task with reward probs that flip in blocks."""

  def __init__(
      self,
      reward_prob_0: float = 0.8,
      reward_prob_1: float = 0.2,
      counterfactual: bool = False,
      **kwargs,
  ):
    
    super().__init__()
    
    # Assign the input parameters as properties
    self.reward_probs = [reward_prob_0, reward_prob_1]
    self._counterfactual = counterfactual
    
  def step(self, choice: int = None):
    """Step the model forward given chosen action."""

    # Sample a reward with this probability
    reward = np.array([float(np.random.binomial(1, prob)) for prob in self.reward_probs], dtype=float)

    # Return the reward
    choice_onehot = np.eye(self.n_actions)[choice]
    if self._counterfactual:
      return reward
    else:
      return choice_onehot * reward[choice] + (1-choice_onehot)*-1

  @property
  def n_actions(self) -> int:
    return 2


class BanditSession(NamedTuple):
  """Holds data for a single session of a bandit task."""
  choices: np.ndarray
  rewards: np.ndarray
  session: np.ndarray
  # reward_probabilities: np.ndarray
  # q: np.ndarray
  n_trials: int
  
  def set_session(self, session: int):
    return self(
      choices=self.choices, 
      rewards=self.rewards, 
      session=np.full_like(self.session, session), 
      # reward_probabilities=self.reward_probabilities, 
      # q=self.q, 
      n_trials=self.n_trials,
      )
  
  def __getitem__(self, val):
    return self._replace(
      choices=self.choices.__getitem__(val), 
      rewards=self.rewards.__getitem__(val), 
      session=self.session.__getitem__(val), 
      # reward_probabilities=self.reward_probabilities.__getitem__(val), 
      # q=self.q.__getitem__(val), 
      n_trials=self.choices.__getitem__(val).shape[0],
      )


###############
#  EXECUTION  #
###############


def run_experiment(
  agent: Agent,
  environment: Bandits,
  n_trials: int,
  session_id: int = 0,
  ) -> BanditSession:
  """Runs a behavioral session from a given agent and environment.

  Args:
    agent: An agent object
    environment: An environment object
    n_trials: The number of steps in the session you'd like to generate

  Returns:
    experiment: A BanditSession holding choices and rewards from the session
  """
  
  choices = np.zeros((n_trials+1)) - 1
  rewards = np.zeros((n_trials+1, agent._n_actions)) - 1
  # qs = np.zeros((n_trials+1, agent._n_actions, agent._state['x_value_reward'].shape[-1])) - 1
  # reward_probs = np.zeros((n_trials+1, agent._n_actions)) - 1

  for trial in range(n_trials+1):
    # Log environment reward probabilities and Q-Values
    # reward_probs[trial] = environment.reward_probs
    # qs[trial] = agent.q
    # First - agent makes a choice
    choice = agent.get_choice()
    # Second - environment computes a reward
    reward = environment.step(choice)
    # Log choice and reward
    choices[trial] = choice
    rewards[trial, :len(reward)] = reward
    
    # Third - agent updates its believes based on chosen action and received reward
    agent.update(choice=choice, reward=rewards[trial], participant_id=session_id)
    
  experiment = BanditSession(n_trials=n_trials,
                             choices=choices[:-1].astype(int),
                             rewards=rewards[:-1],
                             session=np.full(rewards[:-1].shape[0], session_id).astype(int),
                            #  reward_probabilities=reward_probs[:-1],
                            #  q=qs[:-1],
                             )
  return experiment, choices.astype(int), rewards


def create_dataset(
  agent: Agent,
  environment: Bandits,
  n_trials: int,
  n_sessions: int,
  sequence_length: int = None,
  stride: int = 1,
  sample_parameters: bool = False,
  device=torch.device('cpu'),
  verbose=False,
  ) -> tuple[DatasetRNN, list[BanditSession], list[dict[str, float]]]:
  """Generates a behavioral dataset from a given agent and environment.

  Args:
    agent: An agent object to generate choices
    environment: An environment object to generate rewards
    n_trials_per_session: The number of trials in each behavioral session to
      be generated
    n_sessions: The number of sessions to generate
    batch_size: The size of the batches to serve from the dataset. If None, 
      batch_size defaults to n_sessions

  Returns:
    A torch.utils.data.Dataset object suitable for training the RNN object.
    An experliment_list with the results of (simulated) experiments
  """
  
  agent_original = agent
  n_actions = agent[0]._n_actions if isinstance(agent_original, list) else agent._n_actions
  xs = np.zeros((n_sessions, n_trials, n_actions*2 + 1))
  ys = np.zeros((n_sessions, n_trials, n_actions))
  experiment_list = []
  parameter_list = []
  
  print('Creating dataset...')
  for session in tqdm(range(n_sessions)):
    if verbose:
      print(f'Running session {session+1}/{n_sessions}...')
    environment.new_sess()
    if isinstance(agent_original, list):
      agent = agent_original[session]
    agent.new_sess(sample_parameters=sample_parameters, participant_id=session)
    experiment, choices, rewards = run_experiment(agent, environment, n_trials, session)
    experiment_list.append(experiment)
    
    # one-hot encoding of choices
    choices = np.eye(agent._n_actions)[choices]
    ys[session] = choices[1:]
    xs[session] = np.concatenate((choices[:-1], rewards[:-1], experiment.session[:, None]), axis=-1)
    
    if isinstance(agent, AgentQ):
      # add current parameters to list
      parameter_list.append(
        {
          'beta_reward': copy(agent._betas['x_value_reward']),
          'alpha_reward': copy(agent._alpha_reward),
          'alpha_penalty': copy(agent._alpha_penalty),
          'beta_choice': copy(agent._betas['x_value_choice']),
          'alpha_choice': copy(agent._alpha_choice),
          'forget_rate': copy(agent._forget_rate),
        }
      )

  dataset = DatasetRNN(
    xs=xs, 
    ys=ys,
    sequence_length=sequence_length,
    stride=stride,
    device=device)
  
  return dataset, experiment_list, parameter_list


def get_update_dynamics(experiment: Union[np.ndarray, torch.Tensor], agent: Agent, additional_signals: List[str] = ['x_value_reward', 'x_learing_rate_reward', 'x_value_choice']):
  """Compute Q-Values of a specific agent for a specific experiment sequence with given actions and rewards.

  Args:
      experiment (BanditSession): _description_
      agent (_type_): _description_

  Returns:
      _type_: _description_
  """

  if isinstance(experiment, np.ndarray) or isinstance(experiment, torch.Tensor):
    if isinstance(experiment, torch.Tensor):
      experiment = experiment.detach().cpu().numpy()
    if len(experiment.shape) == 3:
      experiment = experiment.squeeze(0)
    # get number of actual trials
    n_trials = len(experiment) - np.argmax(experiment[::-1][:, 0] != -1)
    choices = experiment[:n_trials, :agent._n_actions]
    rewards = experiment[:n_trials, agent._n_actions:2*agent._n_actions]
    # TODO: additional_inputs are currently treated as signals and as meta-information for the embedding
    additional_inputs = experiment[0, 2*agent._n_actions:-3]
    current_block = int(experiment[0, -3])
    experiment_id = int(experiment[0, -2])
    participant_id = int(experiment[0, -1])
  else:
    raise TypeError("experiment is of not of class numpy.ndarray or torch.Tensor")
  
  # reset agent states according to ID
  agent.new_sess(participant_id=participant_id, experiment_id=experiment_id, additional_embedding_inputs=additional_inputs)
  
  # initialize storages
  q = np.zeros((n_trials, agent._n_actions))
  values_signal = {signal: np.zeros((n_trials, agent._n_actions)) for signal in additional_signals}
  choice_probs = np.zeros((n_trials, agent._n_actions))
  
  for trial in range(n_trials):
    # track all states
    q[trial] = agent.q
    for signal in additional_signals:
      values_signal[signal][trial] = agent.get_state_value(signal) if signal in agent._state else np.zeros_like(agent.q)
    
    choice_probs[trial] = agent.get_choice_probs()
    
    agent.update(
      choice=np.argmax(choices[trial], axis=-1), 
      reward=rewards[trial],  
      block=current_block, 
      additional_inputs=additional_inputs,
      )
  
  return (q[..., :1], values_signal), choice_probs, agent


###############
# DIAGNOSTICS #
###############


def plot_session(
  choices: np.ndarray,
  rewards: np.ndarray,
  timeseries: Tuple[np.ndarray],
  timeseries_name: str,
  labels: Optional[Tuple[str]] = None,
  title: str = '',
  x_label = 'Trials',
  fig_ax = None,
  compare=False,
  color=None,
  x_axis_info=True,
  y_axis_info=True,
  ):
  """Plot data from a single behavioral session of the bandit task.

  Args:
    choices: The choices made by the agent
    rewards: The rewards received by the agent
    timeseries: The dynamical value of interest on each arm
    timeseries_name: The name of the timeseries
    labels: The labels for the lines in the plot
    fig_ax: A tuple of a matplotlib figure and axis to plot on
    compare: If True, plot multiple timeseries on the same plot
    color: A list of colors to use for the plot; at least as long as the number of timeseries
  """

  if color == None:
    color = [None]*len(timeseries)
  
  # Make the plot
  if fig_ax is None:
    fig, ax = plt.subplots(1, figsize=(10, 3))
  else:
    fig, ax = fig_ax
    
  if compare:
    if timeseries.ndim==2:
      timeseries = np.expand_dims(timeseries, -1)
    if timeseries.ndim!=3 or timeseries.shape[-1]!=1:
      raise ValueError('If compare: timeseries must be of shape (agent, timesteps, 1).')
  else:
    if timeseries.ndim!=2:
      raise ValueError('timeseries must be of shape (timesteps, n_actions).')
                       
  if not compare:
    # choices = np.expand_dims(choices, 0)
    timeseries = np.expand_dims(timeseries, 0)
  
  for i in range(timeseries.shape[0]):
    if labels is not None:
      if timeseries[i].ndim == 1:
        timeseries[i] = timeseries[i, :, None]
      if not compare:
        if len(labels) != timeseries[i].shape[1]:
          raise ValueError('labels length must match timeseries.shape[1].')
      else:
        if timeseries[i].shape[1] != 1:
          raise ValueError('If compare: timeseries.shape[1] must be 1.')
        if len(labels) != timeseries.shape[0]:
          raise ValueError('If compare: labels length must match timeseries.shape[0].')
      for ii in range(timeseries[i].shape[-1]):
          label = labels[ii] if not compare else labels[i]
          ax.plot(timeseries[i, :, ii], label=label, color=color[i])
      ax.legend(bbox_to_anchor=(1, 1))
    else:  # Skip legend.
      ax.plot(timeseries[i], color=color[i])
  
  # Plot ticks relating to whether the option was chosen (factual) or not (counterfactual) and whether it was rewarded
  min_y, max_y = np.min(timeseries), np.max(timeseries)
  # diff_min_max = np.max((5e-2, max_y - min_y))  # make sure the difference is not <= 0
  diff_min_max = np.max((1e-1, max_y - min_y))
  
  x = np.arange(len(choices))
  chosen_y = min_y - 1e-1  # Lower position for chosen (bigger tick)
  not_chosen_y = max_y + 1e-1  # Slightly lower for not chosen (smaller tick)
  # not_chosen_y = chosen_y - 1e-1 * diff_min_max  # Slightly lower for not chosen (smaller tick)

  # if (rewards > 0 and rewards < 1).any():
  # Plot ticks for chosen options
  ax.scatter(x[(choices == 1) & (rewards == 1)], np.full(sum((choices == 1) & (rewards == 1)), chosen_y), color='green', s=120, marker='|')  # Large green tick for chosen reward
  ax.scatter(x[(choices == 1) & (rewards == 0)], np.full(sum((choices == 1) & (rewards == 0)), chosen_y), color='red', s=80, marker='|')  # Large red tick for chosen penalty

  # Plot ticks for not chosen options
  ax.scatter(x[(choices == 0) & (rewards == 1)], np.full(sum((choices == 0) & (rewards == 1)), not_chosen_y), color='green', s=120, marker='|')  # Small green tick
  ax.scatter(x[(choices == 0) & (rewards == 0)], np.full(sum((choices == 0) & (rewards == 0)), not_chosen_y), color='red', s=80, marker='|')  # Small red tick
  # else:
  #   # Plot ticks for chosen options
  #   ax.scatter(x[(choices == 1)], np.full(sum((choices == 1)), chosen_y), color='green', s=100, marker='|')  # Large green tick for chosen reward
  #   ax.scatter(x[(choices == 1)], np.full(sum((choices == 1)), chosen_y), color='red', s=80, marker='|')  # Large red tick for chosen penalty

  #   # Plot ticks for not chosen options
  #   ax.scatter(x[(choices == 0) & (rewards == 1)], np.full(sum((choices == 0)), not_chosen_y), color='green', s=100, marker='|')  # Small green tick
  #   ax.scatter(x[(choices == 0) & (rewards == 0)], np.full(sum((choices == 0)), not_chosen_y), color='red', s=80, marker='|')  # Small red tick
  # ax.set_ylim(not_chosen_y, np.max((-not_chosen_y, np.max(timeseries + 1e-1 * diff_min_max))))
  
  if x_axis_info:
    ax.set_xlabel(x_label)
  else:
    # ax.set_xticks(np.linspace(1, len(timeseries), 5))
    ax.set_xticklabels(['']*5)
    
  if y_axis_info:
    ax.set_ylabel(timeseries_name)
  else:
    ax.set_yticks(np.linspace(0, 1, 5))
    ax.set_yticklabels(['']*5)
    
  ax.set_title(title)



from typing import List, Dict, Tuple
import numpy as np
from math import comb

import pysindy as ps

from .sindy_utils import remove_control_features, conditional_filtering
from .rnn_utils import DatasetRNN
from .sindy_utils import generate_off_policy_data, create_dataset
from .bandits import AgentNetwork

def fit_sindy(
    variables: List[np.ndarray], 
    control: List[np.ndarray] = None, 
    rnn_modules: List[str] = None,
    control_signals: List[str] = None, 
    polynomial_degree: int = 1, 
    library_setup: Dict[str, List[str]] = {},
    filter_setup: Dict[str, Tuple[str, float]] = {},
    optimizer_type: str = "SR3_weighted_l1",
    optimizer_threshold: float = 0.05,
    optimizer_alpha: float = 1,
    catch_convergence_warning: bool = False,
    verbose: bool = False,
    ):
    
    # get all x-features
    x_features = rnn_modules
    # get all control features
    c_features = control_signals
    
    # make sure that all x_features are in the library_setup
    for feature in x_features:
        if feature not in library_setup:
            library_setup[feature] = []
    
    # train one sindy model per variable
    sindy_models = {feature: None for feature in x_features}
    loss = 0
    for index_feature, x_feature in enumerate(x_features):
        if verbose:
            print(f'\nSINDy model for {x_feature}:')
        
        # sort signals into corresponding arrays    
        x_i = [x[:, index_feature].reshape(-1, 1) for x in variables] # get current x-feature as target variable
        control_i = control
        
        feature_names_i = [x_feature] + c_features
        
        # filter target variable and control features according to filter conditions
        if x_feature in filter_setup:
            if not isinstance(filter_setup[x_feature][0], list):
                # check that filter_setup[x_feature] is a list of filter-conditions 
                filter_setup[x_feature] = [filter_setup[x_feature]]
            for filter_condition in filter_setup[x_feature]:
                x_i, control_i, feature_names_i = conditional_filtering(
                    x_train=x_i, 
                    control=control_i, 
                    feature_names=feature_names_i, 
                    feature_filter=filter_condition[0], 
                    condition=filter_condition[1], 
                    remove_feature_filter=False,
                )
        
        # remove unnecessary control features according to library setup
        control_i = remove_control_features(control_i, feature_names_i[1:], library_setup[x_feature])
        feature_names_i = [x_feature] + library_setup[x_feature]
        
        # add a dummy control feature if no control features are remaining - otherwise sindy breaks
        # if control_i is None or len(control_i) == 0:
        #     raise NotImplementedError('Having no control signal in a module is currently not implemented')
        #     control_i = None
        #     feature_names_i = feature_names_i + ['dummy']
        
        # Set up increasing thresholds with polynomial degree for SR3_weighted_l1
        if optimizer_type == "SR3_weighted_l1":
            n_polynomial_combinations = np.array([comb(len(feature_names_i) + d, d) for d in range(polynomial_degree+1)])
            thresholds = np.zeros((1, n_polynomial_combinations[-1]))
            index = 0
            for d in range(len(n_polynomial_combinations)):
                thresholds[0, index:n_polynomial_combinations[d]] = d * optimizer_threshold
                index = n_polynomial_combinations[d]

        # Create optimizer based on type
        if optimizer_type == "STLSQ":
            optimizer = ps.STLSQ(alpha=optimizer_alpha, threshold=optimizer_threshold)
        elif optimizer_type == "SR3_L1":
            optimizer = ps.SR3(
                thresholder="L1",
                nu=optimizer_alpha,
                threshold=optimizer_threshold,
                verbose=verbose,
                max_iter=100
            )
        else:  # "SR3_weighted_l1" (default)
            optimizer = ps.SR3(
                thresholder="weighted_l1",
                nu=optimizer_alpha,
                threshold=optimizer_threshold,
                thresholds=thresholds,
                verbose=verbose,
                max_iter=100
            )
            
        # Setup sindy model for current x-feature
        sindy_models[x_feature] = ps.SINDy(
            optimizer=optimizer,
            feature_library=ps.PolynomialLibrary(polynomial_degree, interaction_only=False),
            discrete_time=True,
            feature_names=feature_names_i,
        )

        # fit sindy model
        sindy_models[x_feature].fit(x_i, u=control_i, t=1, multiple_trajectories=True, ensemble=False)
        
        if catch_convergence_warning and sindy_models[x_feature].optimizer.iters >= sindy_models[x_feature].optimizer.max_iter-1:
            raise RuntimeError("SINDy optimizer did not converge.")
        
        # drop very small sindy coefficients for sparsity
        coefficients = sindy_models[x_feature].model.steps[-1][1].coef_[0]
        for i, c in enumerate(coefficients):
            if np.abs(c) < optimizer_threshold:
                sindy_models[x_feature].model.steps[-1][1].coef_[0, i] = 0
            elif np.abs(c-1) < optimizer_threshold:
                sindy_models[x_feature].model.steps[-1][1].coef_[0, i] = 1
        
        if verbose:
            sindy_models[x_feature].print()
    
    return sindy_models


def fit_sindy_pipeline(
    participant_id: int,
    agent: AgentNetwork,
    data: DatasetRNN,
    rnn_modules: List,
    control_signals: List,
    sindy_library_setup: Dict,
    sindy_filter_setup: Dict,
    sindy_dataprocessing: Dict,
    optimizer_type: str,
    optimizer_alpha: float,
    optimizer_threshold: float,
    polynomial_degree: int,
    shuffle: bool,
    n_sessions_off_policy: int,
    n_trials_off_policy: int,
    n_trials_same_action_off_policy: int,
    catch_convergence_warning: bool,
    verbose: bool,
    ):
    
    # get participant data
    if n_sessions_off_policy == 0 and data is not None:
        mask_participant_id = data.xs[:, 0, -1] == participant_id
        dataset_fit = DatasetRNN(*data[mask_participant_id])
    elif n_sessions_off_policy > 0:
        mask_participant_id = (data.xs[:, 0, -1] == participant_id).int().argmax().item()
        dataset_fit = generate_off_policy_data(
            participant_id=participant_id,
            experiment_id=data.xs[mask_participant_id, 0, -2],
            block=data.xs[mask_participant_id, 0, -3],
            additional_inputs=data.xs[mask_participant_id, 0, agent._n_actions*2:-3],
            n_trials_off_policy=n_trials_off_policy,
            n_trials_same_action_off_policy=n_trials_same_action_off_policy,
            n_sessions_off_policy=n_sessions_off_policy,
            n_actions=agent._n_actions,
            )
    else:
        raise ValueError("One of the arguments data or n_sessions_off_policy (> 0) must be given. If n_sessions_off_policy > 0 the SINDy modules will be fitted on the off-policy data regardless of data. If n_sessions_off_policy = 0 then data will be used to fit the SINDy modules.")
    
    # extract all necessary data from the RNN (memory state) and align with the control inputs (action, reward)
    rnn_variables, control_parameters, _, _ = create_dataset(
        agent=agent,
        data=dataset_fit,
        rnn_modules=rnn_modules,
        control_signals=control_signals,
        shuffle=shuffle,
        dataprocessing=sindy_dataprocessing,
    )
    
    # fit one SINDy-model per RNN-module
    sindy_modules = fit_sindy(
        variables=rnn_variables,
        control=control_parameters,
        rnn_modules=rnn_modules,
        control_signals=control_signals,
        polynomial_degree=polynomial_degree,
        library_setup=sindy_library_setup,
        filter_setup=sindy_filter_setup,
        optimizer_type=optimizer_type,
        optimizer_alpha=optimizer_alpha,
        optimizer_threshold=optimizer_threshold,
        catch_convergence_warning=catch_convergence_warning,
        verbose=verbose,
    )
    
    return sindy_modules


import sys
import os
import warnings

import numpy as np
import matplotlib.pyplot as plt
from typing import List, Dict

sys.path.append('resources')
from resources.bandits import AgentQ, BanditsDrift, BanditsSwitch, plot_session, create_dataset as create_dataset_bandits
from resources.rnn import RLRNN_eckstein2022
from resources.rnn_utils import parameter_file_naming
from resources.sindy_utils import check_library_setup, save_spice, SindyConfig_eckstein2022
from resources.sindy_training import fit_spice
from utils.convert_dataset import convert_dataset
from utils.plotting import plot_session as plot_session
from utils.setup_agents import setup_agent_rnn

warnings.filterwarnings("ignore")

def main(
    class_rnn: type = None,
    model: str = None,
    data: str = None,
    save: bool = False,
    additional_inputs_data: List[str] = None,
    
    # generated dataset parameters
    participant_id: int = None,
    
    # sindy config
    rnn_modules: List[str] = None,
    control_parameters: List[str] = None,
    library_setup: Dict[str, list] = None,
    filter_setup: Dict[str, list] = None,
    dataprocessing_setup: Dict[str, list] = None,
    
    # sindy parameters
    optimizer_type: str = "SR3_L1",
    optimizer_threshold = 0.05,
    optimizer_alpha = 0.1,
    polynomial_degree = 2,
    n_trials_off_policy = 1024,
    n_sessions_off_policy = 1,
    n_trials_same_action_off_policy = 5,
    verbose = True,
    use_optuna = False,
    filter_bad_participants = False,  # Added parameter to control filtering
    pruning = False,
    train_test_ratio = 1.0,
    optuna_threshold = 0.03,
    optuna_n_trials = 50,
    
    # ground truth parameters
    beta_reward = 3.,
    alpha = 0.25,
    alpha_penalty = -1.,
    forget_rate = 0.,
    confirmation_bias = 0.,
    beta_choice = 0.,
    alpha_choice = 0.,
    alpha_counterfactual = 0.,
    parameter_variance = 0.,
    
    # environment parameters
    n_actions = 2,
    sigma = .2,
    counterfactual = False,
    
    get_loss: bool = False,
    analysis: bool = False,
    # reward_range: List[float] = [0, 1],
    ):

    # ---------------------------------------------------------------------------------------------------
    # SINDy-agent setup
    # ---------------------------------------------------------------------------------------------------
    
    # # tracked variables and control signals in the RNN
    # rnn_modules = ['x_learning_rate_reward', 'x_value_reward_not_chosen', 'x_value_choice_chosen', 'x_value_choice_not_chosen']
    # control_parameters = ['c_action', 'c_reward_chosen', 'c_value_reward', 'c_value_choice']
    
    # # library setup: 
    # # which terms are allowed as control inputs in each SINDy model
    # # key is the SINDy model name, value is a list of allowed control inputs from the list of control signals 
    # library_setup = {
    #     'x_learning_rate_reward': ['c_reward_chosen', 'c_value_reward', 'c_value_choice'],
    #     'x_value_reward_not_chosen': ['c_reward_chosen', 'c_value_choice'],
    #     'x_value_choice_chosen': ['c_value_reward'],
    #     'x_value_choice_not_chosen': ['c_value_reward'],
    # }

    # # data-filter setup: 
    # # which samples are allowed as training samples in each SINDy model based on the given filter condition (conditions are always equality conditions)
    # # key is the SINDy model name, value is a list with a triplet of values:
    # #   1. str: feature name to be used as a filter
    # #   2. numeric: the numeric filter condition
    # #   3. bool: remove feature from control inputs --> TODO: check if this is necessary or makes things just more complicated
    # # Multiple conditions can also be given as a list of triplets.
    # # Example:
    # #   'x_value_choice_not_chosen': ['c_action', 0, True] means that for the SINDy model 'x_value_choice_not_chosen', only samples where the feature 'c_action' == 0 are used for training the SINDy model. 
    # #   The control parameter 'c_action' is removed afterwards from the list of control signals for training of the model
    # filter_setup = {
    #     # 'x_value_reward_chosen': ['c_action', 1, True], -> Remove this one as well
    #     'x_learning_rate_reward': ['c_action', 1, True],
    #     'x_value_reward_not_chosen': ['c_action', 0, True],
    #     'x_value_choice_chosen': ['c_action', 1, True],
    #     'x_value_choice_not_chosen': ['c_action', 0, True],
    # }

    # # data pre-processing setup:
    # # define the processing steps for each variable and control signal.
    # # possible processing steps are: 
    # #   1. Trimming: Remove the first 25% of the samples along the time-axis. This is useful if the RNN begins with a variable at 0 but then accumulates first first to a specific default value, i.e. the range changes from (0, p) to (q, q+p). That way the data is cleared of the accumulation process. Trimming will be active for all variables, if it is active for one. 
    # #   2. Offset-Clearing: Clearup any offset by determining the minimal value q of a variable and move the value range from (q, q+p) -> (0, p). This step makes SINDy equations less complex and aligns them more with RL-Theory
    # #   3. Normalization: Scale the value range of a variable to x_max - x_min = 1. Offset-Clearing is recommended to achieve a value range of (0, 1) 
    # # The processing steps are passed in the form of a binary triplet in this order: (Trimming, Offset-Clearing, Normalization) 
    # dataprocessing_setup = {
    #     'x_learning_rate_reward': [0, 0, 0],
    #     'x_value_reward_not_chosen': [0, 0, 0],
    #     'x_value_choice_chosen': [1, 1, 0],
    #     'x_value_choice_not_chosen': [1, 1, 0],
    #     # 'c_action': [0, 0, 0],
    #     # 'c_reward': [0, 0, 0],
    #     'c_value_reward': [0, 0, 0],
    #     'c_value_choice': [1, 1, 0],
    # }
    
    sindy_feature_list = rnn_modules + control_parameters

    if not check_library_setup(library_setup, sindy_feature_list, verbose=True):
        raise ValueError('Library setup does not match feature list.')
        
    # ---------------------------------------------------------------------------------------------------
    # RNN Setup
    # ---------------------------------------------------------------------------------------------------
    
    # set up rnn agent and expose q-values to train sindy
    if model is None:
        file_rnn = parameter_file_naming('params/params', beta_reward=beta_reward, alpha_reward=alpha, alpha_penalty=alpha_penalty, beta_choice=beta_choice, alpha_choice=alpha_choice, forget_rate=forget_rate, confirmation_bias=confirmation_bias, alpha_counterfactual=alpha_counterfactual, variance=parameter_variance, verbose=True)
    else:
        file_rnn = model
    
    agent_rnn = setup_agent_rnn(
        class_rnn=class_rnn,
        path_model=file_rnn,
        list_sindy_signals=rnn_modules+control_parameters,
    )
    
    # ---------------------------------------------------------------------------------------------------
    # Data setup
    # ---------------------------------------------------------------------------------------------------
    
    agent = None
    if data is None:
        # set up ground truth agent and environment
        environment = BanditsDrift(sigma=sigma, n_actions=n_actions, counterfactual=counterfactual)
        agent = AgentQ(
            n_actions=n_actions, 
            beta_reward=beta_reward, 
            alpha_reward=alpha, 
            alpha_penalty=alpha_penalty, 
            beta_choice=beta_choice, 
            alpha_choice=alpha_choice, 
            forget_rate=forget_rate, 
            confirmation_bias=confirmation_bias, 
            alpha_counterfactual=alpha_counterfactual,
            )
        
        if participant_id is not None or participant_id != 0:
            participant_id = 0
        participant_ids = np.arange(agent_rnn._model.n_participants, dtype=int)
        dataset, _, _ = create_dataset_bandits(agent, environment, 200, len(participant_ids))
    else:
        # get data from experiments for later evaluation
        dataset, _, df, _ = convert_dataset(data, additional_inputs=additional_inputs_data)
        participant_ids = dataset.xs[..., -1].unique().int().cpu().numpy()

    choices = dataset.xs[..., 0] == 1
    reward_range = [dataset.xs[..., n_actions][choices].min(), dataset.xs[..., n_actions][choices].max()]
    print(reward_range)
    # ---------------------------------------------------------------------------------------------------
    # SINDy training
    # ---------------------------------------------------------------------------------------------------
    
    if use_optuna and verbose:
        print("\nUsing Optuna to find optimal optimizer configuration for each participant")
    
    # setup the SINDy-agent
    agent_spice, loss_spice = fit_spice(
        rnn_modules=rnn_modules,
        control_signals=control_parameters,
        agent_rnn=agent_rnn,
        data=dataset,
        n_sessions_off_policy=n_sessions_off_policy,
        n_trials_off_policy=n_trials_off_policy,
        n_trials_same_action_off_policy=n_trials_same_action_off_policy,
        polynomial_degree=polynomial_degree,
        library_setup=library_setup,
        filter_setup=filter_setup,
        dataprocessing=dataprocessing_setup,
        optimizer_type=optimizer_type,
        optimizer_threshold=optimizer_threshold,
        optimizer_alpha=optimizer_alpha,
        get_loss=get_loss,
        participant_id=participant_id,
        shuffle=False,
        verbose=verbose,
        use_optuna=use_optuna,
        filter_bad_participants=filter_bad_participants,
        pruning=pruning,
        train_test_ratio=train_test_ratio,
        optuna_threshold=optuna_threshold,
        optuna_n_trials=optuna_n_trials,
        )
    
    # If agent_spice is None, we couldn't fit the model, so return early
    if len(participant_ids) == 0:
        print("ERROR: Failed to fit SPICE model. Returning None.")
        return None, None, None
    
    # save spice modules
    if save:
        file_spice = file_rnn.split(os.path.sep)
        if 'rnn' in file_spice[-1]:
            file_spice[-1] = file_spice[-1].replace('rnn', 'spice')
        else:
            file_spice[-1] = 'spice_' + file_spice[-1]
        file_spice = os.path.join(*file_spice)
        save_spice(agent_spice=agent_spice, file=file_spice)
        print("Saved SPICE parameters to file " + file_spice)
    
    # ---------------------------------------------------------------------------------------------------
    # Analysis
    # ---------------------------------------------------------------------------------------------------
    
    if analysis and len(participant_ids) > 0:
        participant_id_test = participant_id if participant_id is not None else participant_ids[0]
        mask_participant_id = dataset.xs[:, 0, -1] == participant_id_test
        experiment_test = dataset.xs[mask_participant_id][0]
        
        agent_rnn.new_sess(participant_id=participant_id_test)
        agent_spice.new_sess(participant_id=participant_id_test)
        
        # print sindy equations from tested sindy agent
        print('\nDiscovered SPICE models:')
        for module in rnn_modules:
            if participant_id_test in agent_spice._model.submodules_sindy[module]:
                agent_spice._model.submodules_sindy[module][participant_id_test].print()
            else:
                print(f"No module available for {module}, participant {participant_id_test}")
        print('\n')
        
        # set up ground truth agent by getting parameters from dataset if specified
        if data is not None and agent is None and analysis and 'mean_beta_reward' in df.columns:
            agent = AgentQ(
                beta_reward = df['beta_reward'].values[(df['session']==participant_id_test).values][0],
                alpha_reward = df['alpha_reward'].values[(df['session']==participant_id_test).values][0],
                alpha_penalty = df['alpha_penalty'].values[(df['session']==participant_id_test).values][0],
                confirmation_bias = df['confirmation_bias'].values[(df['session']==participant_id_test).values][0],
                forget_rate = df['forget_rate'].values[(df['session']==participant_id_test).values][0],
                beta_choice = df['beta_choice'].values[(df['session']==participant_id_test).values][0],
                alpha_choice = df['alpha_choice'].values[(df['session']==participant_id_test).values][0],
            )
        
        # get analysis plot
        if agent is not None:
            agents = {'groundtruth': agent, 'rnn': agent_rnn, 'sindy': agent_spice}
            plt_title = r'$GT:\beta_{reward}=$'+str(np.round(agent._beta_reward, 2)) + r'; $\beta_{choice}=$'+str(np.round(agent._beta_choice, 2))+'\n'
        else:
            agents = {'rnn': agent_rnn, 'sindy': agent_spice}
            plt_title = ''
        
        fig, axs = plot_session(agents, experiment_test, reward_range=reward_range)
        betas = agent_spice.get_betas()
        plt_title += r'SINDy: $\beta_{reward}=$'+str(np.round(betas['x_value_reward'], 2)) + r'; $\beta_{choice}=$'+str(np.round(betas['x_value_choice'], 2))
        
        fig.suptitle(plt_title)
        plt.show()
        # plt.savefig('example_recovered_dynamics.png', dpi=500)
    
    features = {}
    # for module in agent_spice._model.submodules_sindy:
    #     features[module] = {}
    #     for pid in agent_spice._model.submodules_sindy[module]:
    #         features_i = np.array(agent_spice._model.submodules_sindy[module][pid].get_feature_names())
    #         coeffs_i = agent_spice._model.submodules_sindy[module][pid].coefficients()[0]
    #         index_u = [not 'dummy' in f for f in features_i]
    #         features_i = features_i[index_u]
    #         coeffs_i = coeffs_i[index_u]
    #         # features[model][pid] = tuple(features_i)
    #         features[module][pid] = tuple(coeffs_i)
    
    # features['beta_reward'] = {}
    # features['beta_choice'] = {}
    # for pid in participant_ids:
    #     pid = int(pid)
    #     if pid in agent_spice._model.submodules_sindy[rnn_modules[0]]:
    #         agent_spice.new_sess(participant_id=pid, additional_embedding_inputs=)
    #         betas = agent_spice.get_betas()
    #         features['beta_reward'][pid] = betas['x_value_reward']
    #         features['beta_choice'][pid] = betas['x_value_choice']
        
    return agent_spice, features, loss_spice


if __name__=='__main__':
    import argparse
    
    parser = argparse.ArgumentParser(description="Run SINDy pipeline with optimizer selection")
    parser.add_argument("--model", type=str, help="Path to model file")
    parser.add_argument("--data", type=str, default=None, help="Path to data file")
    parser.add_argument("--participant_id", type=int, default=None, help="Participant ID")
    parser.add_argument("--polynomial_degree", type=int, default=2, help="Polynomial degree")
    parser.add_argument("--optimizer_type", type=str, default="SR3_weighted_l1", choices=["STLSQ", "SR3_L1", "SR3_weighted_l1"], help="Optimizer type")
    parser.add_argument("--optimizer_alpha", type=float, default=0.1, help="Optimizer alpha")
    parser.add_argument("--optimizer_threshold", type=float, default=0.05, help="Optimizer threshold")
    parser.add_argument("--optuna_threshold", type=float, default=0.1, help="Threshold to use Optuna: Difference of action probabilities between RNN and discovered SPICE model")
    parser.add_argument("--optuna_n_trials", type=int, default=50, help="Number of Optuna trials to optimize SPICE fitting configuration")
    parser.add_argument("--use_optuna", action="store_true", help="Use Optuna to find the best optimizer for each participant")
    parser.add_argument("--filter_bad_participants", action="store_true", help="Remove badly fitted participants")
    parser.add_argument("--n_trials_off_policy", type=int, default=1000, help="Number of trials for off-policy data")
    parser.add_argument("--n_trials_same_action_off_policy", type=int, default=5, help="Number of same actions in a row for off-policy data")
    parser.add_argument("--n_sessions_off_policy", type=int, default=1, help="Number of sessions for off-policy data")
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument("--analysis", action="store_true", help="Perform analysis")
    parser.add_argument("--get_loss", action="store_true", help="Compute loss")
    parser.add_argument("--save", action="store_true", help="Pickle SPICE parameters")
    parser.add_argument('--train_test_ratio', type=str, default="1.0", help='Ratio of training data; Can also be a comma-separated list of integeres to indicate testing sessions.')
    
    args = parser.parse_args()
    
    # convert train_test_ratio to number of list of numbers
    if ',' in args.train_test_ratio:
        train_test_ratio = [int(x) for x in args.train_test_ratio.split(',')]
    else:
        train_test_ratio = float(args.train_test_ratio)
    
    class_rnn = RLRNN_eckstein2022
    sindy_config = SindyConfig_eckstein2022
    
    agent_spice, features, loss = main(
        class_rnn=class_rnn,
        model=args.model,
        data=args.data,
        save=args.save,
        participant_id=args.participant_id,
        polynomial_degree=args.polynomial_degree,
        optimizer_type=args.optimizer_type,
        optimizer_alpha=args.optimizer_alpha,
        optimizer_threshold=args.optimizer_threshold,
        n_trials_off_policy=args.n_trials_off_policy,
        n_trials_same_action_off_policy=args.n_trials_same_action_off_policy,
        n_sessions_off_policy=args.n_sessions_off_policy,
        verbose=args.verbose,
        analysis=args.analysis,
        get_loss=args.get_loss,
        use_optuna=args.use_optuna,
        filter_bad_participants=args.filter_bad_participants,
        train_test_ratio=args.train_test_ratio,
        optuna_threshold=args.optuna_threshold,
        optuna_n_trials=args.optuna_n_trials,
        **sindy_config,
    )
    
    if loss is not None:
        print(f"Loss: {loss}")


"""
Unit tests for the RNN-SINDy theorist.
"""
import pytest
import numpy as np
import torch
from spice import rnn_sindy_theorist
from spice.resources import BaseRNN


class SimpleRNN(BaseRNN):
    """Simple RNN implementation for testing."""
    def __init__(self, n_actions, hidden_size, dropout, n_participants=0, n_experiments=0):
        super().__init__()
        self.n_actions = n_actions
        self.hidden_size = hidden_size
        self.rnn = torch.nn.RNN(
            input_size=5,  # Fixed for testing
            hidden_size=hidden_size,
            num_layers=1,
            batch_first=True
        )
        self.fc = torch.nn.Linear(hidden_size, n_actions)
        self.dropout = torch.nn.Dropout(dropout)
        
        if n_participants > 0:
            self.participant_embedding = torch.nn.Embedding(n_participants, 1)
        if n_experiments > 0:
            self.experiment_embedding = torch.nn.Embedding(n_experiments, 1)

    def forward(self, x, hidden=None):
        out, hidden = self.rnn(x, hidden)
        out = self.dropout(out)
        out = self.fc(out)
        return out, hidden


def test_theorist_initialization(model_config):
    """Test that the theorist can be properly initialized."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    assert model is not None
    assert isinstance(model.rnn_model, SimpleRNN)
    assert model.hidden_size == model_config['hidden_size']
    assert model.n_actions == model_config['n_actions']


def test_theorist_fit(model_config, sample_data):
    """Test the fit method of the theorist."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    
    model.fit(sample_data['conditions'], sample_data['targets'])
    
    # Check that the model components are properly initialized after fitting
    assert model.rnn_agent is not None
    assert model.sindy_agent is not None
    assert model.sindy_features is not None


def test_theorist_predict(model_config, sample_data):
    """Test the predict method of the theorist."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    
    model.fit(sample_data['conditions'], sample_data['targets'])
    pred_rnn, pred_sindy = model.predict(sample_data['conditions'])
    
    # Check shapes
    expected_shape = (
        sample_data['n_participants'],
        sample_data['n_trials'],
        sample_data['n_actions']
    )
    assert pred_rnn.shape == expected_shape
    assert pred_sindy.shape == expected_shape
    
    # Check that predictions are probabilities
    assert np.all(pred_rnn >= 0) and np.all(pred_rnn <= 1)
    assert np.all(pred_sindy >= 0) and np.all(pred_sindy <= 1)
    assert np.allclose(np.sum(pred_rnn, axis=-1), 1.0, atol=1e-6)
    assert np.allclose(np.sum(pred_sindy, axis=-1), 1.0, atol=1e-6)


def test_get_sindy_features(model_config, sample_data):
    """Test retrieving SINDy features."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    
    # Should raise error before fitting
    with pytest.raises(ValueError):
        model.get_sindy_features()
    
    model.fit(sample_data['conditions'], sample_data['targets'])
    features = model.get_sindy_features()
    
    # Check that features are returned for each participant
    assert isinstance(features, dict)
    assert len(features) == sample_data['n_participants']


def test_get_participant_embeddings(model_config, sample_data):
    """Test retrieving participant embeddings."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    
    model.fit(sample_data['conditions'], sample_data['targets'])
    embeddings = model.get_participant_embeddings()
    
    assert embeddings is not None
    assert len(embeddings) == sample_data['n_participants']
    assert all(isinstance(emb, torch.Tensor) for emb in embeddings.values())


def test_get_experiment_embeddings(model_config, sample_data):
    """Test retrieving experiment embeddings."""
    model = rnn_sindy_theorist(
        rnn_class=SimpleRNN,
        **model_config
    )
    
    model.fit(sample_data['conditions'], sample_data['targets'])
    embeddings = model.get_experiment_embeddings()
    
    assert embeddings is not None
    assert len(embeddings) == model_config['n_experiments']
    assert all(isinstance(emb, torch.Tensor) for emb in embeddings.values()) 


"""
SPICE: Sparse and Interpretable Cognitive Equations
"""

from .estimator import SpiceEstimator, SpiceConfig
from .precoded import (
    RescorlaWagnerRNN,
    ForgettingRNN,
    LearningRateRNN,
    ParticipantEmbeddingRNN,
    Weinhardt2024RNN
)
from .resources.rnn import BaseRNN

__version__ = "0.1.0"
__all__ = [
    "SpiceEstimator",
    "SpiceConfig",
    "RescorlaWagnerRNN",
    "ForgettingRNN",
    "LearningRateRNN",
    "ParticipantEmbeddingRNN",
    "Weinhardt2024RNN",
    "BaseRNN",
] 


import sys
import os

from typing import List, Dict
from torch import device, load
import numpy as np
import torch
import pickle

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from resources.rnn import BaseRNN
from resources.bandits import AgentSpice, AgentNetwork, AgentQ
from resources.sindy_training import fit_spice
from resources.sindy_utils import load_spice
from utils.convert_dataset import convert_dataset


def setup_rnn(
    class_rnn,
    path_model,
    n_actions=2,
    counterfactual=False,
    device=device('cpu'),
    **kwargs,
) -> BaseRNN:
    
    # get n_participants and hidden_size from state dict
    state_dict = torch.load(path_model, map_location=torch.device('cpu'))
    
    if 'model' in state_dict:
        state_dict = state_dict['model']
    
    participant_embedding_index = [i for i, s in enumerate(list(state_dict.keys())) if 'participant_embedding' in s]
    participant_embedding_bool = True if len(participant_embedding_index) > 0 else False
    n_participants = 0 if not participant_embedding_bool else state_dict[list(state_dict.keys())[participant_embedding_index[0]]].shape[0]
    
    key_hidden_size = [key for key in state_dict if 'x' in key.lower()][0]  # first key that contains the hidden_size
    hidden_size = state_dict[key_hidden_size].shape[0]
    
    key_embedding_size = [key for key in state_dict if 'embedding' in key.lower()]
    if len(key_embedding_size) > 0:
        embedding_size = state_dict[key_embedding_size[0]].shape[1]
    else:
        embedding_size = 0
        
    rnn = class_rnn(
        n_actions=n_actions, 
        hidden_size=hidden_size, 
        embedding_size=embedding_size,
        n_participants=n_participants, 
        device=device, 
        counterfactual=counterfactual,
        )
    rnn.load_state_dict(state_dict)
        
    return rnn


def setup_agent_rnn(
    class_rnn,
    path_model,
    n_actions=2,
    counterfactual=False,
    deterministic=True,
    device=device('cpu'),
    **kwargs,
    ) -> AgentNetwork:
    
    rnn = setup_rnn(class_rnn=class_rnn, path_model=path_model, device=device, n_actions=n_actions, counterfactual=counterfactual)
    agent = AgentNetwork(model_rnn=rnn, n_actions=n_actions, deterministic=deterministic)
    
    return agent


def setup_agent_spice(
    class_rnn: type,
    path_rnn: str,
    path_spice: str = None,
    path_data: str = None,
    rnn_modules: List[str] = None,
    control_parameters: List[str] = None,
    sindy_library_polynomial_degree: int = None,
    sindy_library_setup: Dict[str, List] = None,
    sindy_filter_setup: Dict[str, List] = None,
    sindy_dataprocessing: Dict[str, List] = None,
    threshold: float = 0.05,
    regularization: float = 0.1,
    participant_id: int = None,
    deterministic: bool = True,
    filter_bad_participants: bool = False,
    use_optuna: bool = False,
    verbose: bool = False,
    **kwargs,
) -> AgentSpice:
    
    agent_rnn = setup_agent_rnn(class_rnn=class_rnn, path_model=path_rnn)
    
    if path_spice is None or path_spice == '':
        dataset = convert_dataset(file=path_data)[0]

        # fit SPICE model to RNN
        agent_spice, _ = fit_spice(
            agent_rnn=agent_rnn,
            data=dataset,
            rnn_modules=rnn_modules,
            control_signals=control_parameters,
            polynomial_degree=sindy_library_polynomial_degree,
            library_setup=sindy_library_setup,
            filter_setup=sindy_filter_setup,
            dataprocessing=sindy_dataprocessing,
            participant_id=participant_id,
            n_sessions_off_policy=1,
            n_trials_off_policy=1000,
            optimizer_type='SR3_weighted_l1',
            optimizer_alpha=regularization,
            optimizer_threshold=threshold,
            deterministic=deterministic,
            filter_bad_participants=filter_bad_participants,
            use_optuna=use_optuna,
            verbose=verbose,
        )
    else:
        # load SPICE model from a file
        spice_modules = load_spice(path_spice)
        agent_spice = AgentSpice(model_rnn=agent_rnn._model, sindy_modules=spice_modules, n_actions=agent_rnn._n_actions)
        
    return agent_spice



import time
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, RandomSampler
# import tqdm

from .rnn import BaseRNN
from .rnn_utils import DatasetRNN


def gradient_penalty(f: nn.Module, e_i: torch.Tensor, e_j: torch.Tensor, factor=1.0):
    
    # one-hot encode
    e_i = f[0].one_hot_encode(e_i)
    e_j = f[0].one_hot_encode(e_j)
    
    # e_i, e_j: one-hot tensors, shape [batch, P]
    eps = torch.rand(e_i.size(0), 1).to(e_i.device)
    eps = eps.expand_as(e_i)
    
    # Interpolation in participant‐ID space
    e_hat = eps*e_i + (1-eps)*e_j
    e_hat.requires_grad_(True)

    # compute interpolated embedding
    emb_hat = f[1](f[0].get_embedding(e_hat))
    
    # Compute Jacobian norm
    grads = torch.autograd.grad(
        outputs=emb_hat,
        inputs=e_hat,
        grad_outputs=torch.ones_like(emb_hat),
        create_graph=True
    )[0]  # shape [batch, P]
    gp = (grads.view(grads.size(0), -1).norm(2, dim=1) ** 2).mean()
    return factor * gp


class ReduceOnPlateauWithRestarts:
    def __init__(self, optimizer, min_lr, factor, patience):
        """
        Plateau-based LR scheduler with restarts to the base learning rate when min_lr is hit.

        Parameters:
        - optimizer: Optimizer instance.
        - min_lr: The minimum learning rate after reductions.
        - factor: Multiplicative factor to reduce the LR on plateau.
        - patience: Number of epochs with no improvement before reducing the LR.
        """
        self.optimizer = optimizer
        self.min_lr = min_lr
        self.base_lrs = [group['initial_lr'] for group in optimizer.param_groups]  # Extract base LRs
        self.factor = factor
        self.patience = patience
        self.base_patience = patience
        
        self.best = float('inf')  # Initialize the best validation loss as infinity
        self.num_bad_epochs = 0  # Initialize the count of bad epochs
        self.num_cycles_completed = 0
        
        # Store the current learning rate for each parameter group
        self.current_lrs = [group['lr'] for group in optimizer.param_groups]

    def step(self, metrics):
        """
        Update the learning rate based on the validation loss.
        """
        if metrics < self.best:
            self.best = metrics
            self.num_bad_epochs = 0  # Reset bad epochs counter
        else:
            self.num_bad_epochs += 1
        
        # Check if patience is exceeded
        if self.num_bad_epochs > self.patience:
            self._reduce_lr()  # Reduce learning rates
            self._adjust_patience()  # Adjust the patience according to the learning rate
            self.num_bad_epochs = 0  # Reset bad epochs counter

    def _reduce_lr(self):
        """
        Reduce the learning rate for all parameter groups by the given factor.
        """
        for i, param_group in enumerate(self.optimizer.param_groups):            
            
            old_lr = param_group['lr']
            new_lr = max(old_lr * self.factor, self.min_lr)
            
            # Check if the new learning rate has hit min_lr and reset if so
            if new_lr <= self.min_lr:
                param_group['lr'] = self.base_lrs[i]
                self.num_cycles_completed += 1
            else:
                param_group['lr'] = new_lr
    
    def _adjust_patience(self):
        """
        Adjust the patience according to the learning rate.
        """
        # self.patience = max([self.patience * (1+self.num_cycles_completed) if self.get_lr()[-1] < self.base_lrs[-1] else self.base_patience, 200])
        self.patience = self.patience * 2 if self.get_lr()[-1] < self.base_lrs[-1] else self.base_patience

    def get_lr(self):
        """
        Retrieve the current learning rates for all parameter groups.
        """
        return [group['lr'] for group in self.optimizer.param_groups]
    
    def get_last_lr(self):
        """
        Retrieve the last computed learning rates for all parameter groups.
        """
        return [group['lr'] for group in self.optimizer.param_groups]
    

def batch_train(
    model: BaseRNN,
    xs: torch.Tensor,
    ys: torch.Tensor,
    optimizer: torch.optim.Optimizer = None,
    n_steps: int = -1,
    loss_fn: nn.modules.loss._Loss = nn.CrossEntropyLoss(),
    ):

    """
    Trains a model with the given batch.
    """
    
    if n_steps == -1:
        n_steps = xs.shape[1]
        
    model.set_initial_state(batch_size=len(xs))
    state = model.get_state(detach=True)
    
    loss_batch = 0
    iterations = 0
    for t in range(0, xs.shape[1], n_steps):
        n_steps = min(xs.shape[1]-t, n_steps)
        xs_step = xs[:, t:t+n_steps]
        ys_step = ys[:, t:t+n_steps]
        
        mask = xs_step[..., :1] > -1
        
        state = model.get_state(detach=True)
        ys_pred = model(xs_step, state, batch_first=True)[0]
        
        ys_pred = ys_pred * mask
        ys_step = ys_step * mask
        
        loss_step = loss_fn(
            ys_pred.reshape(-1, model._n_actions), 
            torch.argmax(ys_step.reshape(-1, model._n_actions), dim=1),
            )
        
        loss_batch += loss_step
        iterations += 1
        
        if torch.is_grad_enabled():
            
            # backpropagation
            optimizer.zero_grad()
            loss_step.backward()
            optimizer.step()
    
    return model, optimizer, loss_batch.item()/iterations
    

def fit_model(
    model: BaseRNN,
    dataset_train: DatasetRNN,
    dataset_test: DatasetRNN = None,
    optimizer: torch.optim.Optimizer = None,
    convergence_threshold: float = 1e-7,
    epochs: int = 1,
    batch_size: int = -1,
    bagging: bool = False,
    scheduler: bool = False,
    n_steps: int = -1,
    verbose: bool = True,
    path_save_checkpoints: str = None,
    ):
    """_summary_

    Args:
        model (BaseRNN): A child class of the BaseRNN, which implements the forward method
        dataset_train (DatasetRNN): training data for the RNN of shape (Batch, Timesteps, Features) with Features being (Actions, Rewards, Participant ID) -> (n_actions, n_actions, 1)
        dataset_test (DatasetRNN, optional): Validation dataset during training. Defaults to None.
        optimizer (torch.optim.Optimizer, optional): Torch-optimizer. Defaults to None.
        convergence_threshold (float, optional): Threshold of convergence value, which determines early stopping of training. Defaults to 1e-5.
        epochs (int, optional): Total number of training epochs. Defaults to 1.
        batch_size (int, optional): Batch size. Defaults to -1.
        bagging (bool, optional): Enables bootstrap aggregation. Defaults to False.
        n_steps (int, optional): Number of steps passed at once through the RNN to compute a gradient over steps. Defaults to -1.
        verbose (bool, optional): Verbosity. Defaults to True.

    Returns:
        (BaseRNN, Optimizer, float): (Trained RNN, Optimizer with last state, Training loss)
    """
    
    
    # initialize dataloader
    if batch_size == -1:
        batch_size = len(dataset_train)
    
    # use random sampling with replacement
    if bagging:
        batch_size = max(batch_size, 64)
        sampler = RandomSampler(dataset_train, replacement=True, num_samples=batch_size) if bagging else None
    else:
        sampler = None
    dataloader_train = DataLoader(dataset_train, batch_size=batch_size, sampler=sampler, shuffle=True if sampler is None else False)
    if dataset_test is not None:
        dataloader_test = DataLoader(dataset_test, batch_size=len(dataset_test))
    
    # set up learning rate scheduler
    warmup_steps = 1024
    warmup_steps = warmup_steps if epochs > warmup_steps else 1 #int(epochs * 0.125/16)
    if scheduler and optimizer is not None:
        # Define the LambdaLR scheduler for warm-up
        # def warmup_lr_lambda(current_step):
        #     if current_step < warmup_steps / 0.8:
        #         return min(1e-1, float(current_step) / float(max(1, warmup_steps))) / (optimizer.param_groups[0]['lr']) / 100
        #     else:
        #         return 1 - float(current_step) / float(max(1, warmup_steps)) / (optimizer.param_groups[0]['lr']) / 100
        #     return 1.0  # No change after warm-up phase
        default_lr = optimizer.param_groups[0]['lr'] + 0
        def warmup_lr_lambda(current_step):
            scale = 1e-1 / default_lr
            if current_step < warmup_steps * 0.8:
                return 0.1 * scale  # Scaling factor during the first 80% of warmup
            elif current_step < warmup_steps:
                # Linearly anneal towards 1.0 in the last 20% of warmup steps
                progress = (current_step - warmup_steps * 0.8) / (warmup_steps * 0.2)
                return (0.1 + progress * (1.0 - 0.1)) * scale
            else:
                return 1.0  # Default learning rate scaling after warmup

        # Create the scheduler with the Lambda function
        scheduler_warmup = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=warmup_lr_lambda)
                
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer=optimizer, T_0=warmup_steps if warmup_steps > 0 else 64, T_mult=2)
        # scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer=optimizer, factor=0.1, patience=10, threshold=0, cooldown=64, min_lr=1e-6)
        # scheduler = ReduceOnPlateauWithRestarts(optimizer=optimizer, min_lr=1e-6, factor=0.1, patience=8)
    else:
        scheduler_warmup, scheduler = None, None
        
    if epochs == 0:
        continue_training = False
        msg = 'No training epochs specified. Model will not be trained.'
        if verbose:
            print(msg)
    else:
        continue_training = True
        converged = False
        n_calls_to_train_model = 0
        convergence_value = 1
        last_loss = 1
        recency_factor = 0.5
    
    loss_train = 0
    loss_test = 0
    iterations_per_epoch = max(len(dataset_train), 64) // batch_size
    save_at_epoch = warmup_steps
    
    # start training
    while continue_training:
        try:
            loss_train = 0
            loss_test = 0
            t_start = time.time()
            n_calls_to_train_model += 1
            for _ in range(iterations_per_epoch):
                # get next batch
                xs, ys = next(iter(dataloader_train))
                if xs.device != model.device:
                    xs = xs.to(model.device)
                    ys = ys.to(model.device)
                # train model
                model, optimizer, loss_i = batch_train(
                    model=model,
                    xs=xs,
                    ys=ys,
                    optimizer=optimizer,
                    n_steps=n_steps,
                )
                loss_train += loss_i
            loss_train /= iterations_per_epoch
            
            if dataset_test is not None:
                model.eval()
                with torch.no_grad():
                    xs, ys = next(iter(dataloader_test))
                    if xs.device != model.device:
                        xs = xs.to(model.device)
                        ys = ys.to(model.device)
                    # evaluate model
                    _, _, loss_test = batch_train(
                        model=model,
                        xs=xs,
                        ys=ys,
                        optimizer=optimizer,
                    )
                model.train()
            
            # check for convergence
            dloss = last_loss - loss_test if dataset_test is not None else last_loss - loss_train
            convergence_value += recency_factor * (np.abs(dloss) - convergence_value)
            converged = convergence_value < convergence_threshold
            continue_training = not converged and n_calls_to_train_model < epochs
            last_loss = 0
            last_loss += loss_test if dataset_test is not None else loss_train
            
            msg = None
            # if verbose:
            msg = f'Epoch {n_calls_to_train_model}/{epochs} --- L(Train): {loss_train:.7f}'                
            if dataset_test is not None:
                msg += f'; L(Val): {loss_test:.7f}'
            msg += f'; Time: {time.time()-t_start:.2f}s; Convergence: {convergence_value:.2e}'
            if scheduler is not None:
                msg += f'; LR: {scheduler_warmup.get_last_lr()[-1] if n_calls_to_train_model < warmup_steps else scheduler.get_last_lr()[-1]:.2e}'
                if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau) or isinstance(scheduler, ReduceOnPlateauWithRestarts):
                    msg += f"; Metric: {scheduler.best:.7f}; Bad epochs: {scheduler.num_bad_epochs}/{scheduler.patience}"
            if converged:
                msg += '\nModel converged!'
            elif n_calls_to_train_model >= epochs:
                msg += '\nMaximum number of training epochs reached.'
                if not converged:
                    msg += '\nModel did not converge yet.'
                        
            if scheduler is not None:
                if n_calls_to_train_model <= warmup_steps: 
                    scheduler_warmup.step()
                else:
                    if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau) or isinstance(scheduler, ReduceOnPlateauWithRestarts):
                        scheduler.step(metrics=loss_test if dataset_test is not None else loss_train)
                    # elif isinstance(scheduler, torch.optim.lr_scheduler.CosineAnnealingWarmRestarts):
                    #     scheduler.step(epoch=n_calls_to_train_model)
                    else:
                        scheduler.step()
            
            # save checkpoint
            if path_save_checkpoints and n_calls_to_train_model == save_at_epoch:
                torch.save(model.state_dict(), path_save_checkpoints.replace('.', f'_ep{n_calls_to_train_model}.'))
                save_at_epoch *= 2
                
        except KeyboardInterrupt:
            continue_training = False
            msg = 'Training interrupted. Continuing with further operations...'
        
        #if verbose:
        print(msg, end='\r')
            
    return model, optimizer, loss_train


import torch
import torch.nn as nn
from typing import Optional, Tuple, Dict, Iterable, Callable, Union, List
import pysindy as ps
import numpy as np


class GRUModule(nn.Module):
    def __init__(self, input_size, **kwargs):
        super().__init__()
        
        self.gru_in = nn.GRU(input_size, 1)
        self.linear_out = nn.Linear(1, 1)
        
        # Simple weight initialization for all parameters
        self._initialize_weights()
    
    def _initialize_weights(self):
        """Apply Xavier uniform to all parameters"""
        for param in self.parameters():
            if param.dim() > 1:  # Only weight matrices, skip biases
                nn.init.xavier_uniform_(param)
            else:
                nn.init.zeros_(param)
                
    # def _initialize_weights(self):
    #     """Initialize weights using Xavier/Glorot initialization for better training stability"""
    #     # Initialize GRU weights
    #     for name, param in self.gru_in.named_parameters():
    #         if 'weight_ih' in name:  # Input-to-hidden weights
    #             nn.init.xavier_uniform_(param)
    #         elif 'weight_hh' in name:  # Hidden-to-hidden weights
    #             nn.init.orthogonal_(param)
    #         elif 'bias' in name:  # Biases
    #             nn.init.zeros_(param)
    #             # Set forget gate bias to 1 for better gradient flow (if applicable)
    #             n = param.size(0)
    #             param.data[n//3:2*n//3].fill_(1.0)  # GRU reset gate bias
        
    #     # Initialize linear layer weights
    #     nn.init.xavier_uniform_(self.linear_out.weight)
    #     nn.init.zeros_(self.linear_out.bias)

    def forward(self, inputs):
        n_actions = inputs.shape[1]
        inputs = inputs.view(inputs.shape[0]*inputs.shape[1], inputs.shape[2]).unsqueeze(0)
        next_state = self.gru_in(inputs[..., 1:], inputs[..., :1].contiguous())[1].view(-1, n_actions, 1)
        next_state = self.linear_out(next_state)
        return next_state


class DummyModule(nn.Module):
    def __init__(self):
        super().__init__()
        
    def forward(self, inputs: torch.Tensor):
        return torch.ones((*inputs.shape[:-1], 1), device=inputs.device, dtype=torch.float)
    
    
class ParameterModule(nn.Module):
    def __init__(self):
        super().__init__()
        
        self.parameter = nn.Parameter(torch.tensor(1.))
        
    def forward(self, *args, **kwargs):
        return self.parameter
    
    

class SparseLeakyReLU(nn.LeakyReLU):
    
    def __init__(self, negative_slope = 0.01, inplace = False, threshold=0.01):
        """Leaky ReLU which in training-mode behaves like the standard LeakyReLU-module.
        In eval-mode it compares the activation value against a threshold and sets it to 0 if it is below.

        Args:
            negative_slope (float, optional): _description_. Defaults to 0.01.
            inplace (bool, optional): _description_. Defaults to False.
            threshold (float, optional): _description_. Defaults to 0.01.
        """
        super().__init__(negative_slope, inplace)
        self.threshold = threshold
    
    def forward(self, input):
        activation = super().forward(input)
        if self.training:
            return activation
        else:
            if activation < self.threshold:
                return torch.zeros_like(activation)
            else:
                activation
                

class BaseRNN(nn.Module):
    def __init__(
        self, 
        n_actions, 
        n_participants: int = 0,
        embedding_size: int = 0,
        device=torch.device('cpu'),
        ):
        super(BaseRNN, self).__init__()
        
        # define general network parameters
        self.device = device
        self._n_actions = n_actions
        self.embedding_size = embedding_size
        self.n_participants = n_participants
        
        # session recording; used for sindy training; training variables start with 'x' and control parameters with 'c' 
        self.recording = {}
        self.submodules_rnn = nn.ModuleDict()
        self.submodules_eq = dict()
        self.submodules_sindy = dict()
        self.betas = nn.ModuleDict()
        
        self.state = self.set_initial_state()
        
    def forward(self, inputs, prev_state, batch_first=False):
        raise NotImplementedError('This method is not implemented.')
    
    def init_forward_pass(self, inputs, prev_state, batch_first):
        if batch_first:
            inputs = inputs.permute(1, 0, 2)
        
        actions = inputs[:, :, :self._n_actions].float()
        rewards = inputs[:, :, self._n_actions:2*self._n_actions].float()
        additional_inputs = inputs[:, :, 2*self._n_actions:-3].float()
        blocks = inputs[:, :, -3:-2].int().repeat(1, 1, 2)
        experiment_ids = inputs[0, :, -2:-1].int()
        participant_ids = inputs[0, :, -1:].int()
        
        if prev_state is not None:
            self.set_state(prev_state)
        else:
            self.set_initial_state(batch_size=inputs.shape[1])
        
        timesteps = torch.arange(actions.shape[0])
        logits = torch.zeros_like(actions)
        
        return (actions, rewards, blocks, additional_inputs), (participant_ids, experiment_ids), logits, timesteps

    def post_forward_pass(self, logits, batch_first):
        # add model dim again and set state
        # self.set_state(*args)
        
        if batch_first:
            logits = logits.permute(1, 0, 2)
            
        return logits
    
    def set_initial_state(self, batch_size=1):
        """this method initializes the hidden state
        
        Args:
            batch_size (int, optional): batch size. Defaults to 1.

        Returns:
            Tuple[torch.Tensor]: initial hidden state
        """
                
        self.recording = {}
                
        state = {key: torch.full(size=[batch_size, self._n_actions], fill_value=self.init_values[key], dtype=torch.float32, device=self.device) for key in self.init_values}
        
        self.set_state(state)
        return self.get_state()
        
    def set_state(self, state_dict):
        """this method sets the latent variables
        
        Args:
            state (Dict[str, torch.Tensor]): hidden state
        """
        
        # self._state = dict(hidden_habit=habit_state, hidden_value=value_state, habit=habit, value=value)
        self.state = state_dict
      
    def get_state(self, detach=False):
        """this method returns the memory state
        
        Returns:
            Dict[str, torch.Tensor]: Dict of latent variables corresponding to the memory state
        """
        
        state = self.state
        if detach:
            state = {key: state[key].detach() for key in state}

        return state
    
    def to(self, device: torch.device): 
        self.device = device
        super().to(device=device)
        return self
    
    def record_signal(self, key, value: torch.Tensor):
        """appends a new timestep sample to the recording. A timestep sample consists of the value at timestep t-1 and the value at timestep t

        Args:
            key (str): recording key to which append the sample to
            old_value (_type_): value at timestep t-1 of shape (batch_size, feature_dim)
            new_value (_type_): value at timestep t of shape (batch_size, feature_dim)
        """
        
        if key not in self.recording:
            self.recording[key] = []
        self.recording[key].append(value.detach().cpu().numpy())
        
    def get_recording(self, key):
        return self.recording[key]
    
    def setup_constant(self, embedding_size: int = None, leaky_relu: float = 0.01):
        if embedding_size is not None:
            return nn.Sequential(nn.Linear(embedding_size, 1), nn.LeakyReLU(leaky_relu))
        else:
            return ParameterModule()
    
    def setup_embedding(self, num_embeddings: int, embedding_size: int, leaky_relu: float = 0.01, dropout: float = 0.5):
        return torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_size),
            torch.nn.LeakyReLU(leaky_relu),
            torch.nn.Dropout(p=dropout),
            )
    
    def setup_module(self, input_size: int):
        """This method creates the standard RNN-module used in computational discovery of cognitive dynamics

        Args:
            input_size (_type_): The number of inputs (excluding the memory state)
            dropout (_type_): Dropout rate before output layer
            activation (nn.Module, optional): Possibility to include an activation function. Defaults to None.

        Returns:
            torch.nn.Module: A torch module which can be called by one line and returns state update
        """
        
        # GRU network
        module = GRUModule(input_size=input_size)
        
        return module 
    
    def call_module(
        self,
        key_module: str,
        key_state: str,
        action: torch.Tensor = None,
        inputs: Union[torch.Tensor, Tuple[torch.Tensor]] = None,
        participant_embedding: torch.Tensor = None, 
        participant_index: torch.Tensor = None,
        activation_rnn: Callable = None,
        scaling: bool = False,
        ):
        """Used to call a submodule of the RNN. Can be either: 
            1. RNN-module (saved in 'self.submodules_rnn')
            2. SINDy-module (saved in 'self.submodules_sindy')
            3. hard-coded equation (saved in 'self.submodules_eq')

        Args:
            key_module (str): _description_
            key_state (str): _description_
            action (torch.Tensor, optional): _description_. Defaults to None.
            inputs (Union[torch.Tensor, Tuple[torch.Tensor]], optional): _description_. Defaults to None.
            participant_embedding (torch.Tensor, optional): _description_. Defaults to None.
            participant_index (torch.Tensor, optional): _description_. Defaults to None.
            activation_rnn (Callable, optional): _description_. Defaults to None.

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        
        if action is not None:
            action = action.unsqueeze(-1)
        value = self.get_state()[key_state].unsqueeze(-1)
        # value = key_state.unsqueeze(-1)
        
        if inputs is None:
            inputs = torch.zeros((*value.shape[:-1], 0), dtype=torch.float32, device=value.device)
            
        if participant_embedding is None:
            participant_embedding = torch.zeros((*value.shape[:-1], 0), dtype=torch.float32, device=value.device)
        elif participant_embedding.ndim == 2:
            participant_embedding = participant_embedding.unsqueeze(1).repeat(1, value.shape[1], 1)
        
        if isinstance(inputs, tuple):
            inputs = torch.concat([inputs_i.unsqueeze(-1) for inputs_i in inputs], dim=-1)
        elif inputs.dim()==2:
            inputs = inputs.unsqueeze(-1)
        
        if key_module in self.submodules_sindy.keys():                
            # sindy module
            
            # convert to numpy
            value = value.detach().cpu().numpy()
            inputs = inputs.detach().cpu().numpy()
            
            if inputs.shape[-1] == 0:
                # create dummy control inputs
                inputs = torch.zeros((*inputs.shape[:-1], 1))
            
            if participant_index is None:
                participant_index = torch.zeros((1, 1), dtype=torch.int32)
            
            next_value = np.zeros_like(value)
            for index_batch in range(value.shape[0]):
                sindy_model = self.submodules_sindy[key_module][participant_index[index_batch].item()] if isinstance(self.submodules_sindy[key_module], dict) else self.submodules_sindy[key_module]
                next_value[index_batch] = np.concatenate(
                    [sindy_model.predict(value[index_batch, index_action], inputs[index_batch, index_action]) for index_action in range(self._n_actions)], 
                    axis=0,
                    )
            next_value = torch.tensor(next_value, dtype=torch.float32, device=self.device)

        elif key_module in self.submodules_rnn.keys():
            # rnn module
            
            inputs = torch.concat((value, inputs, participant_embedding), dim=-1)
            update_value = self.submodules_rnn[key_module](inputs)
            next_value = value + update_value
            
            if activation_rnn is not None:
                next_value = activation_rnn(next_value)
            
        elif key_module in self.submodules_eq.keys():
            # hard-coded equation
            next_value = self.submodules_eq[key_module](value.squeeze(-1), inputs).unsqueeze(-1)

        else:
            raise ValueError(f'Invalid module key {key_module}.')

        if action is not None:
            # keep only actions necessary for that update and set others to zero
            next_value = next_value * action
        
        # clip next_value to a specific range
        next_value = torch.clip(input=next_value, min=-1e1, max=1e1)
        
        if scaling:
            # scale by inverse noise temperature
            scaling_factor = self.betas[key_state] if isinstance(self.betas[key_state], nn.Parameter) else self.betas[key_state](participant_embedding)
            next_value = next_value * scaling_factor
        
        return next_value.squeeze(-1)
    
    def integrate_sindy(self, modules: Dict[str, Iterable[ps.SINDy]]):
        # check that all provided modules find a place in the RNN
        checked = 0
        for m in modules:
            if m in self.submodules_rnn.keys():
                checked += 1
        assert checked == len(modules), f'Not all provided SINDy modules {tuple(modules.keys())} found a corresponding RNN module or vice versa.\nSINDy integration aborted.'
        
        # replace rnn modules with sindy modules
        self.submodules_sindy = modules
